from autolens import config
from autolens.analysis import galaxy_prior as gp
from autolens.analysis import galaxy as g
from autolens.pixelization import pixelization
from autolens.profiles import mass_profiles, light_profiles
from autolens.autopipe import model_mapper as mm
import pytest
import os


class MockPriorModel:
    def __init__(self, name, cls):
        self.name = name
        self.cls = cls
        self.centre = "centre for {}".format(name)
        self.phi = "phi for {}".format(name)


class MockModelMapper:
    def __init__(self):
        self.classes = {}

    def add_class(self, name, cls):
        self.classes[name] = cls
        return MockPriorModel(name, cls)


class MockModelInstance:
    pass


@pytest.fixture(name='test_config')
def make_test_config():
    return config.DefaultPriorConfig(
        config_folder_path="{}/../{}/priors/default".format(os.path.dirname(os.path.realpath(__file__)),
                                                            "test_files/config"))


@pytest.fixture(name="mapper")
def make_mapper():
    return mm.ModelMapper()


@pytest.fixture(name="galaxy_prior_2")
def make_galaxy_prior_2(mapper, test_config):
    galaxy_prior_2 = gp.GalaxyPrior(variable_redshift=True, light_profile=light_profiles.EllipticalDevVaucouleurs,
                                    mass_profile=mass_profiles.EllipticalCoredIsothermal, config=test_config)
    mapper.galaxy_2 = galaxy_prior_2
    return galaxy_prior_2


@pytest.fixture(name="galaxy_prior")
def make_galaxy_prior(mapper, test_config):
    galaxy_prior_1 = gp.GalaxyPrior(variable_redshift=True, light_profile=light_profiles.EllipticalDevVaucouleurs,
                                    mass_profile=mass_profiles.EllipticalCoredIsothermal, config=test_config)
    mapper.galaxy_1 = galaxy_prior_1
    return galaxy_prior_1


class TestGalaxyPrior:
    def test_init_to_model_mapper(self, mapper, test_config):
        mapper.galaxy_1 = gp.GalaxyPrior(variable_redshift=True, light_profile=light_profiles.EllipticalDevVaucouleurs,
                                         mass_profile=mass_profiles.EllipticalCoredIsothermal, config=test_config)
        assert len(mapper.priors_ordered_by_id) == 13

    def test_multiple_galaxies(self, mapper, test_config):
        mapper.galaxy_1 = gp.GalaxyPrior(variable_redshift=True, light_profile=light_profiles.EllipticalDevVaucouleurs,
                                         mass_profile=mass_profiles.EllipticalCoredIsothermal, config=test_config)
        mapper.galaxy_2 = gp.GalaxyPrior(variable_redshift=True, light_profile=light_profiles.EllipticalDevVaucouleurs,
                                         mass_profile=mass_profiles.EllipticalCoredIsothermal, config=test_config)
        assert len(mapper.prior_models) == 2

    def test_align_centres(self, galaxy_prior, test_config):
        prior_models = galaxy_prior.prior_models

        assert prior_models[0].centre != prior_models[1].centre

        galaxy_prior = gp.GalaxyPrior(variable_redshift=True, light_profile=light_profiles.EllipticalDevVaucouleurs,
                                      mass_profile=mass_profiles.EllipticalCoredIsothermal,
                                      align_centres=True, config=test_config)

        prior_models = galaxy_prior.prior_models

        assert prior_models[0].centre == prior_models[1].centre

    def test_align_phis(self, galaxy_prior, test_config):
        prior_models = galaxy_prior.prior_models

        assert prior_models[0].phi != prior_models[1].phi

        prior_models = gp.GalaxyPrior(variable_redshift=True, light_profile=light_profiles.EllipticalDevVaucouleurs,
                                      mass_profile=mass_profiles.EllipticalCoredIsothermal,
                                      align_orientations=True, config=test_config).prior_models
        assert prior_models[0].phi == prior_models[1].phi


class TestNamedProfiles:
    def test_get_prior_model(self):
        galaxy_prior = gp.GalaxyPrior(variable_redshift=True, light_profile=light_profiles.EllipticalSersic,
                                      mass_profile=mass_profiles.EllipticalSersicMass)

        assert isinstance(galaxy_prior.light_profile, mm.PriorModel)
        assert isinstance(galaxy_prior.mass_profile, mm.PriorModel)

    def test_set_prior_model(self):
        mapper = mm.ModelMapper()
        galaxy_prior = gp.GalaxyPrior(variable_redshift=True, light_profile=light_profiles.EllipticalSersic,
                                      mass_profile=mass_profiles.EllipticalSersicMass)

        mapper.galaxy = galaxy_prior

        assert 15 == len(mapper.priors_ordered_by_id)

        galaxy_prior.light_profile = mm.PriorModel(light_profiles.LightProfile)

        assert 8 == len(mapper.priors_ordered_by_id)


class TestResultForArguments:
    def test_simple_instance_for_arguments(self):
        galaxy_prior = gp.GalaxyPrior(variable_redshift=True, )
        arguments = {galaxy_prior.redshift.redshift: 0.5}
        galaxy = galaxy_prior.instance_for_arguments(arguments)

        assert galaxy.redshift == 0.5

    def test_complicated_instance_for_arguments(self):
        galaxy_prior = gp.GalaxyPrior(variable_redshift=True, align_centres=True,
                                      light_profile=light_profiles.EllipticalSersic,
                                      mass_profile=mass_profiles.SphericalIsothermal)

        arguments = {galaxy_prior.redshift.redshift: 0.5,
                     galaxy_prior.mass_profile.centre.centre_0: 1.0,
                     galaxy_prior.mass_profile.centre.centre_1: 2.0,
                     galaxy_prior.mass_profile.einstein_radius: 3.0,
                     galaxy_prior.light_profile.axis_ratio: 4.0,
                     galaxy_prior.light_profile.phi: 5.0,
                     galaxy_prior.light_profile.intensity: 6.0,
                     galaxy_prior.light_profile.effective_radius: 7.0,
                     galaxy_prior.light_profile.sersic_index: 8.0}

        galaxy = galaxy_prior.instance_for_arguments(arguments)

        assert galaxy.light_profiles[0].centre[0] == 1.0
        assert galaxy.light_profiles[0].centre[1] == 2.0

    def test_gaussian_prior_model_for_arguments(self):
        galaxy_prior = gp.GalaxyPrior(variable_redshift=True, align_centres=True,
                                      light_profile=light_profiles.EllipticalSersic,
                                      mass_profile=mass_profiles.SphericalIsothermal)

        redshift_prior = mm.GaussianPrior(1, 1)
        einstein_radius_prior = mm.GaussianPrior(4, 1)
        intensity_prior = mm.GaussianPrior(7, 1)

        arguments = {galaxy_prior.redshift.redshift: redshift_prior,
                     galaxy_prior.mass_profile.centre.centre_0: mm.GaussianPrior(2, 1),
                     galaxy_prior.mass_profile.centre.centre_1: mm.GaussianPrior(3, 1),
                     galaxy_prior.mass_profile.einstein_radius: einstein_radius_prior,
                     galaxy_prior.light_profile.axis_ratio: mm.GaussianPrior(5, 1),
                     galaxy_prior.light_profile.phi: mm.GaussianPrior(6, 1),
                     galaxy_prior.light_profile.intensity: intensity_prior,
                     galaxy_prior.light_profile.effective_radius: mm.GaussianPrior(8, 1),
                     galaxy_prior.light_profile.sersic_index: mm.GaussianPrior(9, 1)}

        gaussian_galaxy_prior_model = galaxy_prior.gaussian_prior_model_for_arguments(arguments)

        assert gaussian_galaxy_prior_model.redshift.redshift == redshift_prior
        assert gaussian_galaxy_prior_model.mass_profile.einstein_radius == einstein_radius_prior
        assert gaussian_galaxy_prior_model.light_profile.intensity == intensity_prior


class TestPixelization(object):
    def test_pixelization(self, test_config):
        galaxy_prior = gp.GalaxyPrior(variable_redshift=True, pixelization=pixelization.Voronoi,
                                      config=test_config)

        arguments = {galaxy_prior.redshift.redshift: 2.0,
                     galaxy_prior.pixelization.pixels: 10,
                     galaxy_prior.pixelization.regularization_coefficients.regularization_coefficients_0: 5}

        galaxy = galaxy_prior.instance_for_arguments(arguments)

        assert galaxy.pixelization.pixels == 10
        assert galaxy.pixelization.regularization_coefficients == (5,)

    def test_fixed_pixelization(self, test_config):
        galaxy_prior = gp.GalaxyPrior(variable_redshift=True, pixelization=pixelization.Voronoi(),
                                      config=test_config)

        arguments = {galaxy_prior.redshift.redshift: 2.0}

        galaxy = galaxy_prior.instance_for_arguments(arguments)

        assert galaxy.pixelization.pixels == 100
        assert galaxy.pixelization.regularization_coefficients == (1.,)


class TestHyperGalaxy(object):
    def test_hyper_galaxy(self, test_config):
        galaxy_prior = gp.GalaxyPrior(variable_redshift=True, hyper_galaxy=g.HyperGalaxy, config=test_config)

        arguments = {galaxy_prior.redshift.redshift: 2.0,
                     galaxy_prior.hyper_galaxy.contribution_factor: 1,
                     galaxy_prior.hyper_galaxy.noise_factor: 2,
                     galaxy_prior.hyper_galaxy.noise_power: 3}

        galaxy = galaxy_prior.instance_for_arguments(arguments)

        assert galaxy.hyper_galaxy.contribution_factor == 1
        assert galaxy.hyper_galaxy.noise_factor == 2
        assert galaxy.hyper_galaxy.noise_power == 3

    def test_fixed_hyper_galaxy(self, test_config):
        galaxy_prior = gp.GalaxyPrior(variable_redshift=True, hyper_galaxy=g.HyperGalaxy(), config=test_config)

        arguments = {galaxy_prior.redshift.redshift: 2.0}

        galaxy = galaxy_prior.instance_for_arguments(arguments)

        assert galaxy.hyper_galaxy.contribution_factor == 0.
        assert galaxy.hyper_galaxy.noise_factor == 0.
        assert galaxy.hyper_galaxy.noise_power == 1.


class TestFixedProfiles(object):
    def test_fixed_light_property(self):
        galaxy_prior = gp.GalaxyPrior(variable_redshift=True, light_profile=light_profiles.EllipticalSersic())

        assert len(galaxy_prior.constant_light_profiles) == 1

    def test_fixed_light(self):
        galaxy_prior = gp.GalaxyPrior(variable_redshift=True, light_profile=light_profiles.EllipticalSersic())

        arguments = {galaxy_prior.redshift.redshift: 2.0}

        galaxy = galaxy_prior.instance_for_arguments(arguments)

        assert len(galaxy.light_profiles) == 1

    def test_fixed_mass_property(self):
        galaxy_prior = gp.GalaxyPrior(variable_redshift=True, mass_profile=mass_profiles.SphericalNFW())

        assert len(galaxy_prior.constant_mass_profiles) == 1

    def test_fixed_mass(self):
        galaxy_prior = gp.GalaxyPrior(variable_redshift=True, nass_profile=mass_profiles.SphericalNFW())

        arguments = {galaxy_prior.redshift.redshift: 2.0}

        galaxy = galaxy_prior.instance_for_arguments(arguments)

        assert len(galaxy.mass_profiles) == 1

    def test_fixed_and_variable(self):
        galaxy_prior = gp.GalaxyPrior(variable_redshift=True, mass_profile=mass_profiles.SphericalNFW(),
                                      light_profile=light_profiles.EllipticalSersic(),
                                      variable_light=light_profiles.EllipticalSersic)

        arguments = {galaxy_prior.redshift.redshift: 2.0,
                     galaxy_prior.variable_light.axis_ratio: 4.0,
                     galaxy_prior.variable_light.phi: 5.0,
                     galaxy_prior.variable_light.intensity: 6.0,
                     galaxy_prior.variable_light.effective_radius: 7.0,
                     galaxy_prior.variable_light.sersic_index: 8.0,
                     galaxy_prior.variable_light.centre.centre_0: 0,
                     galaxy_prior.variable_light.centre.centre_1: 0}

        galaxy = galaxy_prior.instance_for_arguments(arguments)

        assert len(galaxy.light_profiles) == 2
        assert len(galaxy.mass_profiles) == 1


class TestRedshift(object):
    def test_set_redshift_class(self):
        galaxy_prior = gp.GalaxyPrior(variable_redshift=True, )
        galaxy_prior.redshift = g.Redshift(3)
        assert galaxy_prior.redshift.redshift == 3

    def test_set_redshift_float(self):
        galaxy_prior = gp.GalaxyPrior(variable_redshift=True, )
        galaxy_prior.redshift = 3
        # noinspection PyUnresolvedReferences
        assert galaxy_prior.redshift.redshift == 3

    def test_set_redshift_constant(self):
        galaxy_prior = gp.GalaxyPrior(variable_redshift=True, )
        galaxy_prior.redshift = mm.Constant(3)
        # noinspection PyUnresolvedReferences
        assert galaxy_prior.redshift.redshift == 3


@pytest.fixture(name="galaxy")
def make_galaxy():
    return g.Galaxy(redshift=3, sersic=light_profiles.EllipticalSersic(),
                    exponential=light_profiles.EllipticalExponential(),
                    spherical=mass_profiles.SphericalIsothermal())


class TestFromGalaxy(object):
    def test_redshift(self, galaxy):
        galaxy_prior = gp.GalaxyPrior.from_galaxy(galaxy)

        assert galaxy_prior.redshift.redshift == 3

    def test_profiles(self, galaxy):
        galaxy_prior = gp.GalaxyPrior.from_galaxy(galaxy)

        assert galaxy_prior.sersic == galaxy.sersic
        assert galaxy_prior.exponential == galaxy.exponential
        assert galaxy_prior.spherical == galaxy.spherical

    def test_recover_galaxy(self, galaxy):
        recovered = gp.GalaxyPrior.from_galaxy(galaxy).instance_for_arguments({})

        assert recovered.sersic == galaxy.sersic
        assert recovered.exponential == galaxy.exponential
        assert recovered.spherical == galaxy.spherical
        assert recovered.redshift == galaxy.redshift

    def test_override_argument(self, galaxy):
        recovered = gp.GalaxyPrior.from_galaxy(galaxy)
        assert recovered.hyper_galaxy is None

        recovered = gp.GalaxyPrior.from_galaxy(galaxy, hyper_galaxy=g.HyperGalaxy)
        assert recovered.hyper_galaxy is not None



from autolens.imaging import scaled_array
from autolens import exc
import numpy as np
from functools import wraps
import inspect

import logging

logging.basicConfig()
logger = logging.getLogger(__name__)


class Memoizer(object):
    def __init__(self):
        """
        Class to store the results of a function given a set of inputs.
        """
        self.results = {}
        self.calls = 0
        self.arg_names = None

    def __call__(self, func):
        """
        Memoize decorator. Any time a function is called that a memoizer has been attached to its results are stored in
        the results dictionary or retrieved from the dictionary if the function has already been called with those
        arguments.

        Note that the same memoizer persists over all instances of a class. Any state for a given instance that is not
        given in the representation of that instance will be ignored. That is, it is possible that the memoizer will
        give incorrect results if instance state does not affect __str__ but does affect the value returned by the
        memoized method.

        Parameters
        ----------
        func: function
            A function for which results should be memoized

        Returns
        -------
        decorated: function
            A function that memoizes results
        """
        if self.arg_names is not None:
            raise AssertionError("Instantiate a new Memoizer for each function")
        self.arg_names = inspect.getfullargspec(func).args

        @wraps(func)
        def wrapper(*args, **kwargs):
            key = ", ".join(
                ["('{}', {})".format(arg_name, arg) for arg_name, arg in
                 list(zip(self.arg_names, args)) + [(k, v) for k, v in kwargs.items()]])
            if key not in self.results:
                self.calls += 1
            self.results[key] = func(*args, **kwargs)
            return self.results[key]

        return wrapper


class Mask(scaled_array.ScaledArray):
    """
    A mask represented by an ndarray where True is masked.
    """

    @classmethod
    def empty_for_shape_arc_seconds_and_pixel_scale(cls, shape_arc_seconds, pixel_scale):
        return cls(np.full(tuple(map(lambda d: int(d / pixel_scale), shape_arc_seconds)), True), pixel_scale)

    @classmethod
    def circular(cls, shape_arc_seconds, pixel_scale, radius_mask, centre=(0., 0.)):
        """
        Setup the mask as a circle, using a specified arc second radius.

        Parameters
        ----------
        shape_arc_seconds: (float, float)
            The (x,y) image_shape
        pixel_scale: float
            The arc-second to pixel conversion factor of each pixel.
        radius_mask : float
            The radius of the circular mask in arc seconds.
        centre: (float, float)
            The centre of the mask.
        """

        grid = Mask.empty_for_shape_arc_seconds_and_pixel_scale(shape_arc_seconds, pixel_scale)

        def fill_grid(x, y):
            x_arcsec, y_arcsec = grid.pixel_coordinates_to_arc_second_coordinates((x, y))

            x_arcsec -= centre[0]
            y_arcsec -= centre[1]

            radius_arcsec = np.sqrt(x_arcsec ** 2 + y_arcsec ** 2)

            grid[x, y] = radius_arcsec > radius_mask

        grid.map(fill_grid)

        return cls(grid, pixel_scale)

    @classmethod
    def annular(cls, shape_arc_seconds, pixel_scale, inner_radius, outer_radius, centre=(0., 0.)):
        """
        Setup the mask as a circle, using a specified inner and outer radius in arc seconds.

        Parameters
        ----------
        shape_arc_seconds : (float, float)
            The (x,y) image_shape of the mask
        pixel_scale: float
            The arc-second to pixel conversion factor of each pixel.
        inner_radius : float
            The inner radius of the annulus mask in arc seconds.
        outer_radius : float
            The outer radius of the annulus mask in arc seconds.
        centre: (float, float)
            The centre of the mask.
        """

        grid = Mask.empty_for_shape_arc_seconds_and_pixel_scale(shape_arc_seconds, pixel_scale)

        def fill_grid(x, y):
            x_arcsec, y_arcsec = grid.pixel_coordinates_to_arc_second_coordinates((x, y))

            x_arcsec -= centre[0]
            y_arcsec -= centre[1]

            radius_arcsec = np.sqrt(x_arcsec ** 2 + y_arcsec ** 2)

            grid[x, y] = radius_arcsec > outer_radius or radius_arcsec < inner_radius

        grid.map(fill_grid)

        return cls(grid, pixel_scale)

    @classmethod
    def unmasked(cls, shape_arc_seconds, pixel_scale):
        """
        Setup the mask such that all values are unmasked, thus corresponding to the entire image.

        Parameters
        ----------
        shape_arc_seconds : (float, float)
            The (x,y) image_shape of the mask
        pixel_scale: float
            The arc-second to pixel conversion factor of each pixel.
        """
        grid = Mask.empty_for_shape_arc_seconds_and_pixel_scale(shape_arc_seconds, pixel_scale)
        return cls(np.ma.make_mask_none(grid.shape), pixel_scale)

    @classmethod
    def for_simulate(cls, shape_arc_seconds, pixel_scale, psf_size):

        if psf_size[0] % 2 == 0 or psf_size[1] % 2 == 0 or psf_size[0] != psf_size[1]:
            raise exc.KernelException("PSF Kernel must be odd and square")

        ma = cls.unmasked(shape_arc_seconds, pixel_scale)
        return ma.pad(new_dimensions=(ma.shape[0] + psf_size[0] - 1, ma.shape[1] + psf_size[1] - 1), pad_value=1)

    @property
    def pixels_in_mask(self):
        return int(np.size(self) - np.sum(self))

    @property
    def coordinate_grid(self):
        """
        Compute the image grid_coords grids from a mask, using the center of every unmasked pixel.
        """
        coordinates = self.grid_coordinates

        pixels = self.pixels_in_mask

        grid = np.zeros(shape=(pixels, 2))
        pixel_count = 0

        for x in range(self.shape[0]):
            for y in range(self.shape[1]):
                if not self[x, y]:
                    grid[pixel_count, :] = coordinates[x, y]
                    pixel_count += 1

        return ImageGrid(grid)

    def map_to_1d(self, grid_data):
        """Compute a data grid, which represents the data values of a data-set (e.g. an image, noise, in the mask.

        Parameters
        ----------
        grid_data: ndarray | float | None

        """
        if grid_data is None or isinstance(grid_data, float):
            return grid_data

        pixels = self.pixels_in_mask

        grid = np.zeros(shape=pixels)
        pixel_count = 0

        for x in range(self.shape[0]):
            for y in range(self.shape[1]):
                if not self[x, y]:
                    grid[pixel_count] = grid_data[x, y]
                    pixel_count += 1

        return grid

    def grid_to_pixel(self):
        """
        Compute the mapping of every pixel in the mask to its 2D pixel coordinates.
        """
        pixels = self.pixels_in_mask

        grid = np.zeros(shape=(pixels, 2), dtype='int')
        pixel_count = 0

        for x in range(self.shape[0]):
            for y in range(self.shape[1]):
                if not self[x, y]:
                    grid[pixel_count, :] = x, y
                    pixel_count += 1

        return grid

    def __getitem__(self, coords):
        try:
            return super(Mask, self).__getitem__(coords)
        except IndexError:
            return True

    @property
    def border_pixel_indices(self):
        """Compute the border image data_to_pixels from a mask, where a border pixel is a pixel inside the mask but on
        its edge, therefore neighboring a pixel with a *True* value.
        """

        border_pixels = np.empty(0, dtype='int')
        image_pixel_index = 0

        for x in range(self.shape[0]):
            for y in range(self.shape[1]):
                if not self[x, y]:
                    if self[x + 1, y] or self[x - 1, y] or self[x, y + 1] or \
                            self[x, y - 1] or self[x + 1, y + 1] or self[x + 1, y - 1] \
                            or self[x - 1, y + 1] or self[x - 1, y - 1]:
                        border_pixels = np.append(border_pixels, image_pixel_index)

                    image_pixel_index += 1

        return border_pixels

    def border_sub_pixel_indices(self, sub_grid_size):
        """Compute the border image data_to_pixels from a mask, where a border pixel is a pixel inside the mask but on
        its edge, therefore neighboring a pixel with a *True* value.
        """

        border_sub_pixels = np.empty(0, dtype='int')
        image_pixel_index = 0

        for x in range(self.shape[0]):
            for y in range(self.shape[1]):
                if not self[x, y]:
                    if self[x + 1, y] or self[x - 1, y] or self[x, y + 1] or \
                            self[x, y - 1] or self[x + 1, y + 1] or self[x + 1, y - 1] \
                            or self[x - 1, y + 1] or self[x - 1, y - 1]:

                        x_arcsec, y_arcsec = self.pixel_coordinates_to_arc_second_coordinates((x, y))
                        sub_grid = np.zeros((sub_grid_size ** 2, 2))
                        sub_pixel_count = 0

                        for x1 in range(sub_grid_size):
                            for y1 in range(sub_grid_size):
                                sub_grid[sub_pixel_count, 0] = self.sub_pixel_to_coordinate(x1, x_arcsec, sub_grid_size)
                                sub_grid[sub_pixel_count, 1] = self.sub_pixel_to_coordinate(y1, y_arcsec, sub_grid_size)
                                sub_pixel_count += 1

                        sub_grid_radii = np.add(np.square(sub_grid[:, 0]), np.square(sub_grid[:, 1]))
                        border_sub_pixel_index = image_pixel_index * (sub_grid_size ** 2) + np.argmax(sub_grid_radii)
                        border_sub_pixels = np.append(border_sub_pixels, border_sub_pixel_index)

                    image_pixel_index += 1

        return border_sub_pixels

    @Memoizer()
    def blurring_mask_for_kernel_shape(self, kernel_shape):
        """Compute the blurring mask, which represents all data_to_pixels not in the mask but close enough to it that a
        fraction of their light will be blurring in the image.

        Parameters
        ----------
        kernel_shape : (int, int)
           The sub_grid_size of the psf which defines the blurring region (e.g. the shape of the PSF)
        """

        if kernel_shape[0] % 2 == 0 or kernel_shape[1] % 2 == 0:
            raise exc.MaskException("psf_size of exterior region must be odd")

        blurring_mask = np.full(self.shape, True)

        def fill_grid(x, y):
            if not self[x, y]:
                for y1 in range((-kernel_shape[1] + 1) // 2, (kernel_shape[1] + 1) // 2):
                    for x1 in range((-kernel_shape[0] + 1) // 2, (kernel_shape[0] + 1) // 2):
                        if 0 <= x + x1 <= self.shape[0] - 1 and 0 <= y + y1 <= self.shape[1] - 1:
                            if self[x + x1, y + y1]:
                                blurring_mask[x + x1, y + y1] = False
                        else:
                            raise exc.MaskException(
                                "setup_blurring_mask extends beyond the sub_grid_size of the mask - pad the image"
                                "before masking")

        self.map(fill_grid)

        return Mask(blurring_mask, self.pixel_scale)

    def map_to_2d(self, data):
        """Use mapper to map an input data-set from a *GridData* to its original 2D image.
        Parameters
        -----------
        data : ndarray
            The grid-data which is mapped to its 2D image.
        """
        data_2d = np.zeros(self.shape)

        for (i, pixel) in enumerate(self.grid_to_pixel()):
            data_2d[pixel[0], pixel[1]] = data[i]

        return data_2d


class SparseMask(Mask):
    def __new__(cls, mask, sparse_grid_size, *args, **kwargs):
        sparse_mask = np.full(mask.shape, True)

        for x in range(mask.shape[0]):
            for y in range(mask.shape[1]):
                if not mask[x, y]:
                    if x % sparse_grid_size == 0 and y % sparse_grid_size == 0:
                        sparse_mask[x, y] = False

        return np.array(sparse_mask).view(cls)

    def __init__(self, mask, sparse_grid_size):
        super().__init__(mask)
        self.mask = mask
        self.sparse_grid_size = sparse_grid_size

    @property
    @Memoizer()
    def index_image(self):
        """
        Setup an image which, for each *False* entry in the sparse mask, puts the sparse pixel index in that pixel.

         This is used for computing the image_to_cluster vector, whereby each image pixel is paired to the sparse
         pixel in this image via a neighbor search."""

        sparse_index_2d = np.zeros(self.shape, dtype=int)
        sparse_pixel_index = 0

        for x in range(self.shape[0]):
            for y in range(self.shape[1]):
                if not self[x, y]:
                    sparse_pixel_index += 1
                    sparse_index_2d[x, y] = sparse_pixel_index

        return sparse_index_2d

    @property
    @Memoizer()
    def sparse_to_image(self):
        """
        Compute the mapping of each sparse image pixel to its closest image pixel, defined using a mask of image \
        data_to_pixels.

        Returns
        -------
        cluster_to_image : ndarray
            The mapping between every sparse clustering image pixel and image pixel, where each entry gives the 1D index
            of the image pixel in the self.
        """
        sparse_to_image = np.empty(0, dtype=int)
        image_pixel_index = 0

        for x in range(self.shape[0]):
            for y in range(self.shape[1]):

                if not self[x, y]:
                    sparse_to_image = np.append(sparse_to_image, image_pixel_index)

                if not self.mask[x, y]:
                    image_pixel_index += 1

        return sparse_to_image

    @property
    @Memoizer()
    def image_to_sparse(self):
        """Compute the mapping between every image pixel in the mask and its closest sparse clustering pixel.

        This is performed by going to each image pixel in the *mask*, and pairing it with its nearest neighboring pixel
        in the *sparse_mask*. The index of the *sparse_mask* pixel is drawn from the *sparse_index_image*. This
        neighbor search continue grows larger and larger around a pixel, until a pixel contained in the *sparse_mask* is
        successfully found.

        Returns
        -------
        image_to_cluster : ndarray
            The mapping between every image pixel and its closest sparse clustering pixel, where each entry give the 1D
            index of the sparse pixel in sparse_pixel arrays.

        """
        image_to_sparse = np.empty(0, dtype=int)

        for x in range(self.shape[0]):
            for y in range(self.shape[1]):
                if not self.mask[x, y]:
                    iboarder = 0
                    pixel_match = False
                    while not pixel_match:
                        for x1 in range(x - iboarder, x + iboarder + 1):
                            for y1 in range(y - iboarder, y + iboarder + 1):
                                if 0 <= x1 < self.shape[0] and 0 <= y1 < self.shape[1]:
                                    if not self[x1, y1] and not pixel_match:
                                        image_to_sparse = np.append(image_to_sparse, self.index_image[x1, y1] - 1)
                                        pixel_match = True

                        iboarder += 1
                        if iboarder == 100:
                            raise exc.MaskException('compute_image_to_sparse - Stuck in infinite loop')

        return image_to_sparse


class ImageGrid(np.ndarray):
    """Abstract class for a regular grid of coordinates. On a regular grid, each pixel's arc-second coordinates \
    are represented by the value at the centre of the pixel.

    Coordinates are defined from the top-left corner, such that data_to_image in the top-left corner of an \
    image (e.g. [0,0]) have a negative x-value and positive y-value in arc seconds. The image pixel indexes are \
    also counted from the top-left.

    A regular *grid_coords* is a NumPy array of image_shape [image_pixels, 2]. Therefore, the first element maps \
    to the image pixel index, and second element to its (x,y) arc second coordinates. For example, the value \
    [3,1] gives the 4th image pixel's y coordinate.

    Below is a visual illustration of a regular grid, where a total of 10 data_to_image are unmasked and therefore \
    included in the grid.

    |x|x|x|x|x|x|x|x|x|x|
    |x|x|x|x|x|x|x|x|x|x|     This is an example image.Mask, where:
    |x|x|x|x|x|x|x|x|x|x|
    |x|x|x|x|o|o|x|x|x|x|     x = True (Pixel is masked and excluded from analysis)
    |x|x|x|o|o|o|o|x|x|x|     o = False (Pixel is not masked and included in analysis)
    |x|x|x|o|o|o|o|x|x|x|
    |x|x|x|x|x|x|x|x|x|x|
    |x|x|x|x|x|x|x|x|x|x|
    |x|x|x|x|x|x|x|x|x|x|
    |x|x|x|x|x|x|x|x|x|x|

    This image pixel index's will come out like this (and the direction of arc-second coordinates is highlighted
    around the image.

    pixel_scale = 1.0"

    <--- -ve  x  +ve -->

    |x|x|x|x|x|x|x|x|x|x|  ^   grid_coords[0] = [-0.5,  1.5]
    |x|x|x|x|x|x|x|x|x|x|  |   grid_coords[1] = [ 0.5,  1.5]
    |x|x|x|x|x|x|x|x|x|x|  |   grid_coords[2] = [-1.5,  0.5]
    |x|x|x|x|0|1|x|x|x|x| +ve  grid_coords[3] = [-0.5,  0.5]
    |x|x|x|2|3|4|5|x|x|x|  y   grid_coords[4] = [ 0.5,  0.5]
    |x|x|x|6|7|8|9|x|x|x| -ve  grid_coords[5] = [ 1.5,  0.5]
    |x|x|x|x|x|x|x|x|x|x|  |   grid_coords[6] = [-1.5, -0.5]
    |x|x|x|x|x|x|x|x|x|x|  |   grid_coords[7] = [-0.5, -0.5]
    |x|x|x|x|x|x|x|x|x|x| \/   grid_coords[8] = [ 0.5, -0.5]
    |x|x|x|x|x|x|x|x|x|x|      grid_coords[9] = [ 1.5, -0.5]
    """

    @property
    def no_pixels(self):
        return self.shape[0]

    def __new__(cls, arr, *args, **kwargs):
        return arr.view(cls)

    def __reduce__(self):
        # Get the parent's __reduce__ tuple
        pickled_state = super(ImageGrid, self).__reduce__()
        # Create our own tuple to pass to __setstate__
        class_dict = {}
        for key, value in self.__dict__.items():
            class_dict[key] = value
        new_state = pickled_state[2] + (class_dict,)
        # Return a tuple that replaces the parent's __setstate__ tuple with our own
        return (pickled_state[0], pickled_state[1], new_state)

    def __setstate__(self, state):

        for key, value in state[-1].items():
            setattr(self, key, value)
        super(ImageGrid, self).__setstate__(state[0:-1])


class SubGrid(ImageGrid):
    """Abstract class for a sub of coordinates. On a sub-grid, each pixel is sub-gridded into a uniform grid of
     sub-coordinates, which are used to perform over-sampling in the lens analysis.

    Coordinates are defined from the top-left corner, such that data_to_image in the top-left corner of an
    image (e.g. [0,0]) have a negative x-value and positive y-value in arc seconds. The image pixel indexes are
    also counted from the top-left.

    A sub *grid_coords* is a NumPy array of image_shape [image_pixels, sub_grid_pixels, 2]. Therefore, the first
    element maps to the image pixel index, the second element to the sub-pixel index and third element to that
    sub pixel's (x,y) arc second coordinates. For example, the value [3, 6, 1] gives the 4th image pixel's
    7th sub-pixel's y coordinate.

    Below is a visual illustration of a sub grid. Like the regular grid, the indexing of each sub-pixel goes from
    the top-left corner. In contrast to the regular grid above, our illustration below restricts the mask to just
    2 data_to_image, to keep the illustration brief.

    |x|x|x|x|x|x|x|x|x|x|
    |x|x|x|x|x|x|x|x|x|x|     This is an example image.Mask, where:
    |x|x|x|x|x|x|x|x|x|x|
    |x|x|x|x|x|x|x|x|x|x|     x = True (Pixel is masked and excluded from analysis)
    |x|x|x|x|o|o|x|x|x|x|     o = False (Pixel is not masked and included in analysis)
    |x|x|x|x|x|x|x|x|x|x|
    |x|x|x|x|x|x|x|x|x|x|
    |x|x|x|x|x|x|x|x|x|x|
    |x|x|x|x|x|x|x|x|x|x|
    |x|x|x|x|x|x|x|x|x|x|

    Our regular-grid looks like it did before:

    pixel_scale = 1.0"

    <--- -ve  x  +ve -->

    |x|x|x|x|x|x|x|x|x|x|  ^
    |x|x|x|x|x|x|x|x|x|x|  |
    |x|x|x|x|x|x|x|x|x|x|  |
    |x|x|x|x|x|x|x|x|x|x| +ve  grid_coords[0] = [-1.5,  0.5]
    |x|x|x|0|1|x|x|x|x|x|  y   grid_coords[1] = [-0.5,  0.5]
    |x|x|x|x|x|x|x|x|x|x| -ve
    |x|x|x|x|x|x|x|x|x|x|  |
    |x|x|x|x|x|x|x|x|x|x|  |
    |x|x|x|x|x|x|x|x|x|x| \/
    |x|x|x|x|x|x|x|x|x|x|

    However, we now go to each image-pixel and derive a sub-pixel grid for it. For example, for pixel 0,
    if *sub_grid_size=2*, we use a 2x2 sub-grid:

    Pixel 0 - (2x2):

           grid_coords[0,0] = [-1.66, 0.66]
    |0|1|  grid_coords[0,1] = [-1.33, 0.66]
    |2|3|  grid_coords[0,2] = [-1.66, 0.33]
           grid_coords[0,3] = [-1.33, 0.33]

    Now, we'd normally sub-grid all data_to_image using the same *sub_grid_size*, but for this illustration lets
    pretend we used a sub_grid_size of 3x3 for pixel 1:

             grid_coords[0,0] = [-0.75, 0.75]
             grid_coords[0,1] = [-0.5,  0.75]
             grid_coords[0,2] = [-0.25, 0.75]
    |0|1|2|  grid_coords[0,3] = [-0.75,  0.5]
    |3|4|5|  grid_coords[0,4] = [-0.5,   0.5]
    |6|7|8|  grid_coords[0,5] = [-0.25,  0.5]
             grid_coords[0,6] = [-0.75, 0.25]
             grid_coords[0,7] = [-0.5,  0.25]
             grid_coords[0,8] = [-0.25, 0.25]

    """

    def __init__(self, array, mask, sub_grid_size=1):
        # noinspection PyArgumentList
        super(SubGrid, self).__init__()
        self.sub_grid_size = sub_grid_size
        self.sub_grid_length = int(sub_grid_size ** 2.0)
        self.sub_grid_fraction = 1.0 / self.sub_grid_length
        self.mask = mask

    def __array_finalize__(self, obj):
        if isinstance(obj, SubGrid):
            self.sub_grid_size = obj.sub_grid_size
            self.sub_grid_length = obj.sub_grid_length
            self.sub_grid_fraction = obj.sub_grid_fraction
            self.mask = obj.mask

    @classmethod
    def from_mask(cls, mask, sub_grid_size=1):
        sub_pixel_count = 0

        grid = np.zeros(shape=(mask.pixels_in_mask * sub_grid_size ** 2, 2))

        for x in range(mask.shape[0]):
            for y in range(mask.shape[1]):
                if not mask[x, y]:
                    x_arcsec, y_arcsec = mask.pixel_coordinates_to_arc_second_coordinates((x, y))

                    for x1 in range(sub_grid_size):
                        for y1 in range(sub_grid_size):
                            grid[sub_pixel_count, 0] = mask.sub_pixel_to_coordinate(x1, x_arcsec, sub_grid_size)

                            grid[sub_pixel_count, 1] = mask.sub_pixel_to_coordinate(y1, y_arcsec, sub_grid_size)

                            sub_pixel_count += 1
        return SubGrid(grid, mask, sub_grid_size=sub_grid_size)

    def sub_data_to_image(self, data):
        return np.multiply(self.sub_grid_fraction, data.reshape(-1, self.sub_grid_length).sum(axis=1))

    @property
    @Memoizer()
    def sub_to_image(self):
        """ Compute the pairing of every sub-pixel to its original image pixel from a mask. """
        sub_to_image = np.zeros(shape=(self.mask.pixels_in_mask * self.sub_grid_size ** 2,), dtype='int')
        image_pixel_count = 0
        sub_pixel_count = 0

        for x in range(self.mask.shape[0]):
            for y in range(self.mask.shape[1]):
                if not self.mask[x, y]:
                    for x1 in range(self.sub_grid_size):
                        for y1 in range(self.sub_grid_size):
                            sub_to_image[sub_pixel_count] = image_pixel_count
                            sub_pixel_count += 1

                    image_pixel_count += 1

        return sub_to_image


class GridCollection(object):

    def __init__(self, image, sub, blurring):
        """
        A collection of grids which contain the coordinates of an image. This includes the image's regular grid,
        sub-grid, blurring region, etc. Coordinate grids are passed through the ray-tracing module to set up the image,
        lens and source planes.

        Parameters
        -----------
        image : GridCoordsImage
            A grid of coordinates for the regular image grid.
        sub : GridCoordsImageSub
            A grid of coordinates for the sub-gridded image grid.
        blurring : GridCoordsBlurring
            A grid of coordinates for the blurring regions.
        """
        self.image = image
        self.sub = sub
        self.blurring = blurring

    @classmethod
    def from_mask_sub_grid_size_and_blurring_shape(cls, mask, sub_grid_size, blurring_shape):
        image_coords = mask.coordinate_grid
        sub_grid_coords = SubGrid.from_mask(mask, sub_grid_size)
        blurring_coords = mask.blurring_mask_for_kernel_shape(blurring_shape).coordinate_grid
        return GridCollection(image_coords, sub_grid_coords, blurring_coords)

    def apply_function(self, func):
        return GridCollection(func(self.image), func(self.sub), func(self.blurring))

    def map_function(self, func, *arg_lists):
        return GridCollection(*[func(*args) for args in zip(self, *arg_lists)])

    @property
    def sub_pixels(self):
        return self.sub.shape[0]

    def __getitem__(self, item):
        return [self.image, self.sub, self.blurring][item]


class ImageGridBorder(np.ndarray):

    @property
    def no_pixels(self):
        return self.shape[0]

    def __new__(cls, arr, polynomial_degree=3, centre=(0.0, 0.0), *args, **kwargs):
        border = arr.view(cls)
        border.polynomial_degree = polynomial_degree
        border.centre = centre
        return border

    @classmethod
    def from_mask(cls, mask, polynomial_degree=3, centre=(0.0, 0.0)):
        return cls(mask.border_pixel_indices, polynomial_degree, centre)

    def __reduce__(self):
        # Get the parent's __reduce__ tuple
        pickled_state = super(ImageGridBorder, self).__reduce__()
        # Create our own tuple to pass to __setstate__
        class_dict = {}
        for key, value in self.__dict__.items():
            class_dict[key] = value
        new_state = pickled_state[2] + (class_dict,)
        # Return a tuple that replaces the parent's __setstate__ tuple with our own
        return pickled_state[0], pickled_state[1], new_state

    def __setstate__(self, state):

        for key, value in state[-1].items():
            setattr(self, key, value)
        super(ImageGridBorder, self).__setstate__(state[0:-1])

    def grid_to_radii(self, grid):
        """
        Convert coordinates to a circular radius.

        If the coordinates have not been transformed to the profile's geometry, this is performed automatically.

        Parameters
        ----------
        grid

        Returns
        -------
        The radius at those coordinates
        """

        return np.sqrt(np.add(np.square(np.subtract(grid[:, 0], self.centre[0])),
                              np.square(np.subtract(grid[:, 1], self.centre[1]))))

    def grid_to_thetas(self, grid):
        """
        Compute the angle in degrees between the image_grid and plane positive x-axis, defined counter-clockwise.

        Parameters
        ----------
        grid : Union((float, float), ndarray)
            The x and y image_grid of the plane.

        Returns
        ----------
        The angle between the image_grid and the x-axis.
        """
        shifted_grid = np.subtract(grid, self.centre)
        theta_from_x = np.degrees(np.arctan2(shifted_grid[:, 1], shifted_grid[:, 0]))
        theta_from_x[theta_from_x < 0.0] += 360.
        return theta_from_x

    def polynomial_fit_to_border(self, grid):

        border_grid = grid[self]

        return np.polyfit(self.grid_to_thetas(border_grid), self.grid_to_radii(border_grid), self.polynomial_degree)

    def move_factors_from_grid(self, grid):
        """Get the move factor of a coordinate.
         A move-factor defines how far a coordinate outside the source-plane setup_border_pixels must be moved in order
         to lie on it. PlaneCoordinates already within the setup_border_pixels return a move-factor of 1.0, signifying
         they are already within the setup_border_pixels.

        Parameters
        ----------
        grid : ndarray
            The x and y image_grid of the pixel to have its move-factor computed.
        """
        grid_thetas = self.grid_to_thetas(grid)
        grid_radii = self.grid_to_radii(grid)
        poly = self.polynomial_fit_to_border(grid)

        with np.errstate(divide='ignore'):
            move_factors = np.divide(np.polyval(poly, grid_thetas), grid_radii)
        move_factors[move_factors > 1.0] = 1.0

        return move_factors

    def relocated_grid_from_grid(self, grid):
        move_factors = self.move_factors_from_grid(grid)
        return np.multiply(grid, move_factors[:, None])


class SubGridBorder(ImageGridBorder):

    @classmethod
    def from_mask(cls, mask, sub_grid_size, polynomial_degree=3, centre=(0.0, 0.0)):
        return cls(mask.border_sub_pixel_indices(sub_grid_size), polynomial_degree, centre)


class BorderCollection(object):

    def __init__(self, image, sub):
        self.image = image
        self.sub = sub

    @classmethod
    def from_mask_and_sub_grid_size(cls, mask, sub_grid_size, polynomial_degree=3, centre=(0.0, 0.0)):
        image_border = ImageGridBorder.from_mask(mask, polynomial_degree, centre)
        sub_border = SubGridBorder.from_mask(mask, sub_grid_size, polynomial_degree, centre)
        return BorderCollection(image_border, sub_border)

    def relocated_grids_from_grids(self, grids):
        return GridCollection(image=self.image.relocated_grid_from_grid(grids.image),
                              sub=self.sub.relocated_grid_from_grid(grids.sub),
                              blurring=None)



from profiling import profiling_data
from profiling import tools
from profiles import geometry_profiles
import numpy as np

class SphericalProfile(geometry_profiles.Profile):

    def __init__(self, centre=(0.0, 0.0)):
        """ Generic elliptical profiles class to contain functions shared by light and mass profiles.

        Parameters
        ----------
        centre: (float, float)
            The coordinates of the centre of the profiles
        axis_ratio : float
            Ratio of profiles ellipse's minor and major axes (b/a)
        phi : float
            Rotational angle of profiles ellipse counter-clockwise from positive x-axis
        """
        super(SphericalProfile, self).__init__(centre)
        self.axis_ratio = 1.0
        self.phi = 0.0

    def grid_angle_to_profile(self, theta_grid):
        return np.cos(theta_grid), np.sin(theta_grid)

    def grid_radius_to_cartesian(self, grid, radius):
        theta_grid = np.arctan2(grid[:, 1], grid[:, 0])
        cos_theta, sin_theta = self.grid_angle_to_profile(theta_grid)
        return np.multiply(radius[:, None], np.vstack((cos_theta, sin_theta)).T)

geometry = SphericalProfile(centre=(0.0, 0.0))

sub_grid_size=4

lsst = profiling_data.setup_class(name='LSST', pixel_scale=0.2, sub_grid_size=sub_grid_size)
euclid = profiling_data.setup_class(name='Euclid', pixel_scale=0.1, sub_grid_size=sub_grid_size)
hst = profiling_data.setup_class(name='HST', pixel_scale=0.05, sub_grid_size=sub_grid_size)
hst_up = profiling_data.setup_class(name='HSTup', pixel_scale=0.03, sub_grid_size=sub_grid_size)
ao = profiling_data.setup_class(name='AO', pixel_scale=0.01, sub_grid_size=sub_grid_size)

lsst_radius = np.ones(lsst.grids.sub.shape[0])
euclid_radius = np.ones(euclid.grids.sub.shape[0])
hst_radius = np.ones(hst.grids.sub.shape[0])
hst_up_radius = np.ones(hst_up.grids.sub.shape[0])
ao_radius = np.ones(ao.grids.sub.shape[0])

@tools.tick_toc_x20
def lsst_solution():
    geometry.grid_radius_to_cartesian(grid=lsst.grids.sub, radius=lsst_radius)

@tools.tick_toc_x20
def euclid_solution():
    geometry.grid_radius_to_cartesian(grid=euclid.grids.sub, radius=euclid_radius)

@tools.tick_toc_x20
def hst_solution():
    geometry.grid_radius_to_cartesian(grid=hst.grids.sub, radius=hst_radius)

@tools.tick_toc_x20
def hst_up_solution():
    geometry.grid_radius_to_cartesian(grid=hst_up.grids.sub, radius=hst_up_radius)

@tools.tick_toc_x20
def ao_solution():
    geometry.grid_radius_to_cartesian(grid=ao.grids.sub, radius=ao_radius)

if __name__ == "__main__":
    lsst_solution()
    euclid_solution()
    hst_solution()
    hst_up_solution()
    ao_solution()


import numpy as np

from autolens.profiles import light_profiles
from profiling import profiling_data
from imaging import convolution
from profiling import tools
import numba
import pytest


class FrameMakerOriginal(object):
    def __init__(self, mask):
        """
        Class to create number array and frames used in 1D convolution
        Parameters
        ----------
        mask: Mask
                A mask where True eliminates weighted_data
        """
        self.mask = mask

        self.number_array = np.full(self.mask.shape, -1)

        count = 0

        for x in range(self.mask.shape[0]):
            for y in range(self.mask.shape[1]):
                if not self.mask[x, y]:
                    self.number_array[x, y] = count
                    count += 1

    def make_frame_array(self, kernel_shape):
        """
        Parameters
        ----------
            An array in which non-masked elements have been numbered 0, 1, 2,...N
        kernel_shape: (int, int)
            The shape of the psf for which frames will be used
        Returns
        -------
        image_frame_indexes: [ndarray]
            A list of frames where the position of a frame corresponds to the number at the centre of that frame
        """
        if kernel_shape[0] % 2 == 0 or kernel_shape[1] % 2 == 0:
            raise exc.KernelException("Kernel must be odd")
        frame_array = []

        # TODO: How would I avoid using for-loops here?
        for x in range(self.number_array.shape[0]):
            for y in range(self.number_array.shape[1]):
                if not self.mask[x][y]:
                    frame_array.append(self.frame_at_coords((x, y), kernel_shape))

        return frame_array

    def make_blurring_frame_array(self, kernel_shape, blurring_region_mask=None):
        """
        Parameters
        ----------
        kernel_shape: (int, int)
            The shape of the psf
        blurring_region_mask: Mask
            A mask describing the boundary of the region from which values may be convoluted

        Returns
        -------
        blurring_frame_indexes [ndarray]
            A list of frames where the position corresponds to a position in the blurring region weighted_data grid and the
            entries correspond to positions in the primary weighted_data grid
        """
        if kernel_shape[0] % 2 == 0 or kernel_shape[1] % 2 == 0:
            raise exc.KernelException("Kernel must be odd")
        if blurring_region_mask is not None and self.mask.shape != blurring_region_mask.shape:
            raise AssertionError("mask and blurring_region_mask must have the same shape")

        frame_array = []
        for x in range(self.mask.shape[0]):
            for y in range(self.mask.shape[1]):
                if self.mask[x][y] and (blurring_region_mask is None or not blurring_region_mask[x, y]):
                    frame = self.frame_at_coords((x, y), kernel_shape)
                    frame_array.append(frame)
        return frame_array

    def frame_at_coords(self, coords, kernel_shape):
        """
        Parameters
        ----------
        coords: (int, int)
            The image_grid of mask_index_array on which the frame should be centred
        kernel_shape: (int, int)
            The shape of the psf for which this frame will be used
        Returns
        -------
        frame: ndarray
            A subset of mask_index_array of shape psf_shape where elements with image_grid outside of image_frame_indexes have
            value -1
        """
        half_x = int(kernel_shape[0] / 2)
        half_y = int(kernel_shape[1] / 2)

        frame = np.full((kernel_shape[0] * kernel_shape[1],), -1)

        for i in range(kernel_shape[0]):
            for j in range(kernel_shape[1]):
                x = coords[0] - half_x + i
                y = coords[1] - half_y + j
                if 0 <= x < self.number_array.shape[0] and 0 <= y < self.number_array.shape[1]:
                    value = self.number_array[x, y]
                    if value >= 0 and not self.mask[x, y]:
                        frame[j + kernel_shape[1] * i] = value

        return frame

    def convolver_for_kernel_shape(self, kernel_shape, blurring_region_mask=None):
        """
        Create a convolver_image that can be used to apply a psf of any shape to a 1D vector of non-masked values
        Parameters
        ----------
        blurring_region_mask: Mask
            A mask describing the blurring region. If False then that pixel is included int he blurring region.
        kernel_shape: (int, int)
            The shape of the psf
        Returns
        -------
            convolver_image: Convolver
        """
        return Convolver(self.make_frame_array(kernel_shape),
                         self.make_blurring_frame_array(
                             kernel_shape,
                             blurring_region_mask) if blurring_region_mask is not None else None)

    def convolver_for_kernel(self, psf):
        return self.convolver_for_kernel_shape(kernel_shape=psf.shape,
                                               blurring_region_mask=self.mask.blurring_mask_for_kernel_shape(
                                                   kernel_shape=psf.shape)).convolver_for_kernel(kernel=psf)


class ConvolverOriginal(object):
    def __init__(self, frame_array, blurring_frame_array):
        """
        Class to convolve a psf with a 1D vector of non-masked values
        Parameters
        ----------
        blurring_frame_array: [ndarray]
            An array of frames created by the frame maker. Maps positions in the psf to values in the 1D vector for
            masked pixels.
        frame_array: [ndarray]
            An array of frames created by the frame maker. A frame maps positions in the psf to values in the 1D
            vector.
        """
        self.frame_array = frame_array
        self.blurring_frame_array = blurring_frame_array

    def convolver_for_kernel(self, kernel):
        """
        Parameters
        ----------
        kernel: ndarray
            An array representing a psf

        Returns
        -------
        convolver_image: KernelConvolver
            An object used to convolve images
        """
        return KernelConvolver(kernel, self.frame_array, self.blurring_frame_array)


class KernelConvolverOriginal(object):

    def __init__(self, kernel, frame_array, blurring_frame_array=None):
        self.shape = kernel.shape

        self.length = self.shape[0] * self.shape[1]
        self.kernel = kernel.flatten()
        self.frame_array = frame_array
        self.blurring_frame_array = blurring_frame_array

    def convolve_array(self, pixel_array, blurring_array=None, sub_shape=None):
        """
        Parameters
        ----------
        blurring_array: [Float]
            An array representing the mapping of a source pixel to a set of image pixels within the blurring region.
        sub_shape: (int, int)
            Defines a sub_grid-region of the psf for which the result should be calculated
        pixel_array: [float]
            A 1D array
        Returns
        -------
        convolved_vector: [float]
            A vector convolved with the psf
        """

        new_array = np.zeros(pixel_array.shape)

        for pixel_index in range(len(pixel_array)):
            frame = self.frame_array[pixel_index]
            value = pixel_array[pixel_index]

            if value > 0:
                new_array = self.convolution_for_value_frame_and_new_array(value, frame, new_array, sub_shape)

        if blurring_array is not None:
            for pixel_index in range(len(blurring_array)):
                frame = self.blurring_frame_array[pixel_index]
                value = blurring_array[pixel_index]

                if value > 0:
                    new_array = self.convolution_for_value_frame_and_new_array(value, frame, new_array, sub_shape)

        return new_array

    def convolve_array_jitted(self, pixel_array, blurring_array=None):
        return self.convolve_array_jit(pixel_array, self.frame_array, blurring_array, self.blurring_frame_array,
                                       self.length, self.kernel)

    @staticmethod
    @numba.jit(nopython=True)
    def convolve_array_jit(pixel_array, frame_array, blurring_array, blurring_frame_array, length, kernel):

        new_array = np.zeros(pixel_array.shape)

        for pixel_index in range(len(pixel_array)):

            frame = frame_array[pixel_index]
            value = pixel_array[pixel_index]

            for kernel_index in range(length):

                vector_index = frame[kernel_index]

                if vector_index == -1:
                    continue
                result = value * kernel[kernel_index]
                if result > 0:
                    new_array[vector_index] += result

        for pixel_index in range(len(blurring_array)):

            frame = blurring_frame_array[pixel_index]
            value = blurring_array[pixel_index]

            for kernel_index in range(length):

                vector_index = frame[kernel_index]

                if vector_index == -1:
                    continue
                result = value * kernel[kernel_index]
                if result > 0:
                    new_array[vector_index] += result

        return new_array

    def convolve_mapping_matrix(self, mapping):
        """
        Simple version of function that applies this convolver_image to a whole mapping matrix.

        Parameters
        ----------
        blurring_array: [Float]
            An array representing the mapping of a source pixel to a set of image pixels within the blurring region.
        array: [float]
            An array representing the mapping of a source pixel to a set of image pixels.

        Returns
        -------
        convolved_array: [float]
            A matrix representing the mapping of source data_to_image to image_grid data_to_image accounting for
            convolution
        """
        blurred_mapping = np.zeros(mapping.shape)
        for i in range(mapping.shape[1]):
            blurred_mapping[:, i] = self.convolve_array(mapping[:, i])
        return blurred_mapping

    def convolve_mapping_matrix_jit(self, mapping):
        """
        Simple version of function that applies this convolver_image to a whole mapping matrix.

        Parameters
        ----------
        blurring_array: [Float]
            An array representing the mapping of a source pixel to a set of image pixels within the blurring region.
        array: [float]
            An array representing the mapping of a source pixel to a set of image pixels.

        Returns
        -------
        convolved_array: [float]
            A matrix representing the mapping of source data_to_image to image_grid data_to_image accounting for
            convolution
        """

        @numba.jit(nopython=True)
        def convolve_matrix_jitted(mapping_matrix, frame_array, length, kernel):

            blurred_mapping = np.zeros(mapping_matrix.shape)

            for pixel_index in range(mapping_matrix.shape[1]):
                for image_index in range(mapping_matrix.shape[0]):

                    frame = frame_array[image_index]
                    value = mapping_matrix[image_index, pixel_index]

                    if value > 0:

                        for kernel_index in range(length):

                            vector_index = frame[kernel_index]

                            if vector_index == -1:
                                continue
                            result = value * kernel[kernel_index]
                            if result > 0:
                                blurred_mapping[vector_index, pixel_index] += result

            return blurred_mapping

        return convolve_matrix_jitted(mapping, self.frame_array, self.length, self.kernel)

    def convolution_for_value_frame_and_new_array(self, value, frame, new_array, sub_shape=None):
        """
        Convolves a value with the psf and populates a new array according to the entries in the frame

        Parameters
        ----------
        value: float
            Some value
        frame: ndarray
            An array describing which entries in the new array convolved values should be inserted into
        new_array: ndarray
            An array into which convolved values are inserted
        sub_shape: (int, int)
            The shape of a reduced sub_grid_size psf

        Returns
        -------
        new_array: ndarray
            The array with values convolved into it
        """

        limits = None
        if sub_shape is not None:
            limits = calculate_limits(self.shape, sub_shape)

        for kernel_index in range(self.length):
            if sub_shape is not None and not is_in_sub_shape(kernel_index, limits, self.shape):
                continue

            vector_index = frame[kernel_index]

            if vector_index == -1:
                continue
            result = value * self.kernel[kernel_index]
            if result > 0:
                new_array[vector_index] += result

        return new_array


sub_grid_size=2
psf_shape = (41, 41)
sersic = light_profiles.EllipticalSersic(centre=(0.0, 0.0), axis_ratio=0.8, phi=90.0, intensity=0.1,
                                         effective_radius=0.8, sersic_index=4.0)

lsst = profiling_data.setup_class(name='LSST', pixel_scale=0.2, sub_grid_size=sub_grid_size, psf_shape=psf_shape)
lsst_kernel_convolver = KernelConvolverOriginal(kernel=lsst.image.psf.trim(psf_shape),
                                                 frame_array=lsst.masked_image.convolver.frame_array,
                                                 blurring_frame_array=lsst.masked_image.convolver.blurring_frame_array)
lsst_image = sersic.intensity_from_grid(grid=lsst.grids.image)
lsst_blurring_image = sersic.intensity_from_grid(grid=lsst.grids.blurring)

assert (lsst_kernel_convolver.convolve_array(lsst_image, lsst_blurring_image) == 
        pytest.approx(lsst_kernel_convolver.convolve_array_jitted(lsst_image, lsst_blurring_image)))

euclid = profiling_data.setup_class(name='Euclid', pixel_scale=0.1, sub_grid_size=sub_grid_size, psf_shape=psf_shape)
euclid_kernel_convolver = KernelConvolverOriginal(kernel=euclid.image.psf.trim(psf_shape),
                                                 frame_array=euclid.masked_image.convolver.frame_array,
                                                 blurring_frame_array=euclid.masked_image.convolver.blurring_frame_array)
euclid_image = sersic.intensity_from_grid(grid=euclid.grids.image)
euclid_blurring_image = sersic.intensity_from_grid(grid=euclid.grids.blurring)
euclid_kernel_convolver.convolve_array_jitted(pixel_array=euclid_image, blurring_array=euclid_blurring_image)

hst = profiling_data.setup_class(name='HST', pixel_scale=0.05, sub_grid_size=sub_grid_size, psf_shape=psf_shape)
hst_kernel_convolver = KernelConvolverOriginal(kernel=hst.image.psf.trim(psf_shape),
                                                 frame_array=hst.masked_image.convolver.frame_array,
                                                 blurring_frame_array=hst.masked_image.convolver.blurring_frame_array)
hst_image = sersic.intensity_from_grid(grid=hst.grids.image)
hst_blurring_image = sersic.intensity_from_grid(grid=hst.grids.blurring)
hst_kernel_convolver.convolve_array_jitted(pixel_array=hst_image, blurring_array=hst_blurring_image)

hst_up = profiling_data.setup_class(name='HSTup', pixel_scale=0.03, sub_grid_size=sub_grid_size, psf_shape=psf_shape)
hst_up_kernel_convolver = KernelConvolverOriginal(kernel=hst_up.image.psf.trim(psf_shape),
                                                 frame_array=hst_up.masked_image.convolver.frame_array,
                                                 blurring_frame_array=hst_up.masked_image.convolver.blurring_frame_array)
hst_up_image = sersic.intensity_from_grid(grid=hst_up.grids.image)
hst_up_blurring_image = sersic.intensity_from_grid(grid=hst_up.grids.blurring)
hst_up_kernel_convolver.convolve_array_jitted(pixel_array=hst_up_image, blurring_array=hst_up_blurring_image)

ao = profiling_data.setup_class(name='AO', pixel_scale=0.01, sub_grid_size=sub_grid_size, psf_shape=psf_shape)
ao_kernel_convolver = KernelConvolverOriginal(kernel=ao.image.psf.trim(psf_shape),
                                                 frame_array=ao.masked_image.convolver.frame_array,
                                                 blurring_frame_array=ao.masked_image.convolver.blurring_frame_array)
ao_image = sersic.intensity_from_grid(grid=ao.grids.image)
ao_blurring_image = sersic.intensity_from_grid(grid=ao.grids.blurring)
ao_kernel_convolver.convolve_array_jitted(pixel_array=ao_image, blurring_array=ao_blurring_image)

@tools.tick_toc_x1
def lsst_solution():
    lsst_kernel_convolver.convolve_array_jitted(pixel_array=lsst_image, blurring_array=lsst_blurring_image)

@tools.tick_toc_x1
def euclid_solution():
    euclid_kernel_convolver.convolve_array_jitted(pixel_array=euclid_image, blurring_array=euclid_blurring_image)

@tools.tick_toc_x1
def hst_solution():
    hst_kernel_convolver.convolve_array_jitted(pixel_array=hst_image, blurring_array=hst_blurring_image)

@tools.tick_toc_x1
def hst_up_solution():
    hst_up_kernel_convolver.convolve_array_jitted(pixel_array=hst_up_image, blurring_array=hst_up_blurring_image)

@tools.tick_toc_x1
def ao_solution():
    ao_kernel_convolver.convolve_array_jitted(pixel_array=ao_image, blurring_array=ao_blurring_image)

if __name__ == "__main__":
    lsst_solution()
    euclid_solution()
    hst_solution()
    hst_up_solution()
    ao_solution()


from autolens import exc
from autolens.analysis import ray_tracing, galaxy
from autolens.profiles import mass_profiles, light_profiles
from astropy import cosmology as cosmo
from autolens.imaging import mask

import pytest
import numpy as np


@pytest.fixture(name="centre_mask")
def make_centre_mask():
    return mask.Mask(np.array([[True, True, True],
                               [True, False, True],
                               [True, True, True]]))


@pytest.fixture(name="grids")
def make_grids(centre_mask):
    grids = mask.GridCollection.from_mask_sub_grid_size_and_blurring_shape(centre_mask, 2, (3, 3))
    grids.image = mask.ImageGrid(np.array([[1.0, 1.0], [1.0, 0.0]]))
    grids.sub = mask.SubGrid(
        np.array([[1.0, 1.0], [1.0, 0.0], [1.0, 1.0], [1.0, 0.0],
                  [1.0, 1.0], [1.0, 0.0], [1.0, 1.0], [1.0, 0.0]]),
        mask=np.array([[True, True, True],
                       [True, False, True],
                       [True, True, True]]), sub_grid_size=2)
    grids.blurring = mask.ImageGrid(np.array([[1.0, 0.0]]))
    return grids


@pytest.fixture(name="no_galaxies")
def make_no_galaxies():
    return [galaxy.Galaxy()]


@pytest.fixture(name='galaxy_no_profiles', scope='function')
def make_galaxy_no_profiles():
    return galaxy.Galaxy()


@pytest.fixture(name="galaxy_light_sersic")
def make_galaxy_light_sersic():
    sersic = light_profiles.EllipticalSersic(axis_ratio=0.5, phi=0.0, intensity=1.0, effective_radius=0.6,
                                             sersic_index=4.0)
    return galaxy.Galaxy(light_profile=sersic)


@pytest.fixture(name="galaxy_mass_sis")
def make_galaxy_mass_sis():
    sis = mass_profiles.SphericalIsothermal(einstein_radius=1.0)
    return galaxy.Galaxy(mass_profile=sis)


@pytest.fixture(name='lens_sis_x3')
def make_lens_sis_x3():
    mass_profile = mass_profiles.SphericalIsothermal(einstein_radius=1.0)
    return galaxy.Galaxy(mass_profile_1=mass_profile, mass_profile_2=mass_profile,
                         mass_profile_3=mass_profile)


@pytest.fixture(name="sparse_mask")
def make_sparse_mask():
    return mask.SparseMask(np.array([[0, 0]]), 1)


@pytest.fixture(name="sparse_mask")
def make_sparse_mask():
    return None


@pytest.fixture(name="galaxy_light_only")
def make_galaxy_light_only():
    return galaxy.Galaxy(light_profile=light_profiles.EllipticalSersic())


@pytest.fixture(name="galaxy_light_and_mass")
def make_galaxy_light_and_mass():
    return galaxy.Galaxy(light_profile=light_profiles.EllipticalSersic(),
                         mass_profile=mass_profiles.SphericalIsothermal())


class MockMassProfile(object):

    def __init__(self, value):
        self.value = value


class MockMapping(object):

    def __init__(self):
        pass


class MockPixelization(object):

    def __init__(self, value):
        self.value = value

    # noinspection PyUnusedLocal,PyShadowingNames
    def reconstructor_from_pix_grids(self, grids, borders, sparse_mask):
        return self.value


class MockBorders(object):

    def __init__(self, image=None, sub=None):
        self.image = image
        self.sub = sub


class TestProperties(object):
    def test_tracer(self, grids):
        tracer = ray_tracing.Tracer([galaxy.Galaxy(hyper_galaxy=galaxy.HyperGalaxy())],
                                    [galaxy.Galaxy(hyper_galaxy=galaxy.HyperGalaxy())],
                                    grids)

        assert tracer.image_plane.hyper_galaxies == [galaxy.HyperGalaxy()]
        assert tracer.source_plane.hyper_galaxies == [galaxy.HyperGalaxy()]

        assert tracer.hyper_galaxies == [galaxy.HyperGalaxy(), galaxy.HyperGalaxy()]

    def test_multi_tracer(self, grids):
        tracer = ray_tracing.MultiTracer([galaxy.Galaxy(hyper_galaxy=galaxy.HyperGalaxy(2), redshift=2),
                                          galaxy.Galaxy(hyper_galaxy=galaxy.HyperGalaxy(1), redshift=1)], grids,
                                         cosmo.Planck15)

        assert tracer.hyper_galaxies == [galaxy.HyperGalaxy(1), galaxy.HyperGalaxy(2)]

    def test_all_with_hyper_galaxies_tracer(self, grids):
        tracer = ray_tracing.Tracer([galaxy.Galaxy(hyper_galaxy=galaxy.HyperGalaxy())],
                                    [galaxy.Galaxy(hyper_galaxy=galaxy.HyperGalaxy())],
                                    grids)

        assert tracer.all_with_hyper_galaxies

        tracer = ray_tracing.Tracer([galaxy.Galaxy(hyper_galaxy=galaxy.HyperGalaxy())],
                                    [galaxy.Galaxy()],
                                    grids)

        assert not tracer.all_with_hyper_galaxies

    def test_all_with_hyper_galaxies_multi_tracer(self, grids):
        tracer = ray_tracing.MultiTracer([galaxy.Galaxy(hyper_galaxy=galaxy.HyperGalaxy(2), redshift=2),
                                          galaxy.Galaxy(hyper_galaxy=galaxy.HyperGalaxy(1), redshift=1)], grids,
                                         cosmo.Planck15)

        assert tracer.all_with_hyper_galaxies

        tracer = ray_tracing.MultiTracer([galaxy.Galaxy(hyper_galaxy=galaxy.HyperGalaxy(2), redshift=2),
                                          galaxy.Galaxy(redshift=1)], grids,
                                         cosmo.Planck15)

        assert not tracer.all_with_hyper_galaxies


class TestTracerGeometry(object):

    def test__2_planes__z01_and_z1(self):
        geometry = ray_tracing.TracerGeometry(redshifts=[0.1, 1.0], cosmology=cosmo.Planck15)

        assert geometry.arcsec_per_kpc(plane_i=0) == pytest.approx(0.525060, 1e-5)
        assert geometry.kpc_per_arcsec(plane_i=0) == pytest.approx(1.904544, 1e-5)

        assert geometry.ang_to_earth(plane_i=0) == pytest.approx(392840, 1e-5)
        assert geometry.ang_between_planes(plane_i=0, plane_j=0) == 0.0
        assert geometry.ang_between_planes(plane_i=0, plane_j=1) == pytest.approx(1481890.4, 1e-5)

        assert geometry.arcsec_per_kpc(plane_i=1) == pytest.approx(0.1214785, 1e-5)
        assert geometry.kpc_per_arcsec(plane_i=1) == pytest.approx(8.231907, 1e-5)

        assert geometry.ang_to_earth(plane_i=1) == pytest.approx(1697952, 1e-5)
        assert geometry.ang_between_planes(plane_i=1, plane_j=0) == pytest.approx(-2694346, 1e-5)
        assert geometry.ang_between_planes(plane_i=1, plane_j=1) == 0.0

        assert geometry.critical_density_kpc(plane_i=0, plane_j=1) == pytest.approx(4.85e9, 1e-2)
        assert geometry.critical_density_arcsec(plane_i=0, plane_j=1) == pytest.approx(17593241668, 1e-2)

    def test__3_planes__z01_z1__and_z2(self):
        geometry = ray_tracing.TracerGeometry(redshifts=[0.1, 1.0, 2.0], cosmology=cosmo.Planck15)

        assert geometry.final_plane == 2

        assert geometry.arcsec_per_kpc(plane_i=0) == pytest.approx(0.525060, 1e-5)
        assert geometry.kpc_per_arcsec(plane_i=0) == pytest.approx(1.904544, 1e-5)

        assert geometry.ang_to_earth(plane_i=0) == pytest.approx(392840, 1e-5)
        assert geometry.ang_between_planes(plane_i=0, plane_j=0) == 0.0
        assert geometry.ang_between_planes(plane_i=0, plane_j=1) == pytest.approx(1481890.4, 1e-5)
        assert geometry.ang_between_planes(plane_i=0, plane_j=2) == pytest.approx(1626471, 1e-5)

        assert geometry.arcsec_per_kpc(plane_i=1) == pytest.approx(0.1214785, 1e-5)
        assert geometry.kpc_per_arcsec(plane_i=1) == pytest.approx(8.231907, 1e-5)

        assert geometry.ang_to_earth(plane_i=1) == pytest.approx(1697952, 1e-5)
        assert geometry.ang_between_planes(plane_i=1, plane_j=0) == pytest.approx(-2694346, 1e-5)
        assert geometry.ang_between_planes(plane_i=1, plane_j=1) == 0.0
        assert geometry.ang_between_planes(plane_i=1, plane_j=2) == pytest.approx(638544, 1e-5)

        assert geometry.arcsec_per_kpc(plane_i=2) == pytest.approx(0.116500, 1e-5)
        assert geometry.kpc_per_arcsec(plane_i=2) == pytest.approx(8.58368, 1e-5)

        assert geometry.ang_to_earth(plane_i=2) == pytest.approx(1770512, 1e-5)
        assert geometry.ang_between_planes(plane_i=2, plane_j=0) == pytest.approx(-4435831, 1e-5)
        assert geometry.ang_between_planes(plane_i=2, plane_j=1) == pytest.approx(-957816)
        assert geometry.ang_between_planes(plane_i=2, plane_j=2) == 0.0

        assert geometry.critical_density_kpc(plane_i=0, plane_j=1) == pytest.approx(4.85e9, 1e-2)
        assert geometry.critical_density_arcsec(plane_i=0, plane_j=1) == pytest.approx(17593241668, 1e-2)

    def test__4_planes__z01_z1_z2_and_z3(self):
        geometry = ray_tracing.TracerGeometry(redshifts=[0.1, 1.0, 2.0, 3.0], cosmology=cosmo.Planck15)

        assert geometry.final_plane == 3

        assert geometry.arcsec_per_kpc(plane_i=0) == pytest.approx(0.525060, 1e-5)
        assert geometry.kpc_per_arcsec(plane_i=0) == pytest.approx(1.904544, 1e-5)

        assert geometry.ang_to_earth(plane_i=0) == pytest.approx(392840, 1e-5)
        assert geometry.ang_between_planes(plane_i=0, plane_j=0) == 0.0
        assert geometry.ang_between_planes(plane_i=0, plane_j=1) == pytest.approx(1481890.4, 1e-5)
        assert geometry.ang_between_planes(plane_i=0, plane_j=2) == pytest.approx(1626471, 1e-5)
        assert geometry.ang_between_planes(plane_i=0, plane_j=3) == pytest.approx(1519417, 1e-5)

        assert geometry.arcsec_per_kpc(plane_i=1) == pytest.approx(0.1214785, 1e-5)
        assert geometry.kpc_per_arcsec(plane_i=1) == pytest.approx(8.231907, 1e-5)

        assert geometry.ang_to_earth(plane_i=1) == pytest.approx(1697952, 1e-5)
        assert geometry.ang_between_planes(plane_i=1, plane_j=0) == pytest.approx(-2694346, 1e-5)
        assert geometry.ang_between_planes(plane_i=1, plane_j=1) == 0.0
        assert geometry.ang_between_planes(plane_i=1, plane_j=2) == pytest.approx(638544, 1e-5)
        assert geometry.ang_between_planes(plane_i=1, plane_j=3) == pytest.approx(778472, 1e-5)

        assert geometry.arcsec_per_kpc(plane_i=2) == pytest.approx(0.116500, 1e-5)
        assert geometry.kpc_per_arcsec(plane_i=2) == pytest.approx(8.58368, 1e-5)

        assert geometry.ang_to_earth(plane_i=2) == pytest.approx(1770512, 1e-5)
        assert geometry.ang_between_planes(plane_i=2, plane_j=0) == pytest.approx(-4435831, 1e-5)
        assert geometry.ang_between_planes(plane_i=2, plane_j=1) == pytest.approx(-957816)
        assert geometry.ang_between_planes(plane_i=2, plane_j=2) == 0.0
        assert geometry.ang_between_planes(plane_i=2, plane_j=3) == pytest.approx(299564)

        assert geometry.arcsec_per_kpc(plane_i=3) == pytest.approx(0.12674, 1e-5)
        assert geometry.kpc_per_arcsec(plane_i=3) == pytest.approx(7.89009, 1e-5)

        assert geometry.ang_to_earth(plane_i=3) == pytest.approx(1627448, 1e-5)
        assert geometry.ang_between_planes(plane_i=3, plane_j=0) == pytest.approx(-5525155, 1e-5)
        assert geometry.ang_between_planes(plane_i=3, plane_j=1) == pytest.approx(-1556945, 1e-5)
        assert geometry.ang_between_planes(plane_i=3, plane_j=2) == pytest.approx(-399419, 1e-5)
        assert geometry.ang_between_planes(plane_i=3, plane_j=3) == 0.0

        assert geometry.critical_density_kpc(plane_i=0, plane_j=1) == pytest.approx(4.85e9, 1e-2)
        assert geometry.critical_density_arcsec(plane_i=0, plane_j=1) == pytest.approx(17593241668, 1e-2)

    def test__scaling_factors__3_planes__z01_z1_and_z2(self):
        geometry = ray_tracing.TracerGeometry(redshifts=[0.1, 1.0, 2.0], cosmology=cosmo.Planck15)

        assert geometry.scaling_factor(plane_i=0, plane_j=1) == pytest.approx(0.9500, 1e-4)
        assert geometry.scaling_factor(plane_i=0, plane_j=2) == pytest.approx(1.0, 1e-4)
        assert geometry.scaling_factor(plane_i=1, plane_j=2) == pytest.approx(1.0, 1e-4)

    def test__scaling_factors__4_planes__z01_z1_z2_and_z3(self):
        geometry = ray_tracing.TracerGeometry(redshifts=[0.1, 1.0, 2.0, 3.0], cosmology=cosmo.Planck15)

        assert geometry.scaling_factor(plane_i=0, plane_j=1) == pytest.approx(0.9348, 1e-4)
        assert geometry.scaling_factor(plane_i=0, plane_j=2) == pytest.approx(0.984, 1e-4)
        assert geometry.scaling_factor(plane_i=0, plane_j=3) == pytest.approx(1.0, 1e-4)
        assert geometry.scaling_factor(plane_i=1, plane_j=2) == pytest.approx(0.754, 1e-4)
        assert geometry.scaling_factor(plane_i=1, plane_j=3) == pytest.approx(1.0, 1e-4)
        assert geometry.scaling_factor(plane_i=2, plane_j=3) == pytest.approx(1.0, 1e-4)


@pytest.fixture(name="light_only_source_plane")
def make_light_only_source_plane(galaxy_light_only, grids):
    return ray_tracing.Plane(galaxies=[galaxy_light_only], grids=grids,
                             compute_deflections=False)


@pytest.fixture(name="light_only_image_plane")
def make_light_only_image_plane(galaxy_light_only, grids):
    return ray_tracing.Plane(galaxies=[galaxy_light_only], grids=grids,
                             compute_deflections=True)


@pytest.fixture(name="light_only_ray_tracer")
def make_light_only_ray_tracer(galaxy_light_only, grids):
    return ray_tracing.Tracer(lens_galaxies=[galaxy_light_only],
                              source_galaxies=[galaxy_light_only],
                              image_plane_grids=grids)


class TestTracer(object):
    class TestSetup:

        def test__image_grid__no_galaxy__image_and_source_planes_setup__same_coordinates(self, grids, no_galaxies):
            ray_trace = ray_tracing.Tracer(lens_galaxies=no_galaxies, source_galaxies=no_galaxies,
                                           image_plane_grids=grids)

            assert ray_trace.image_plane.grids.image[0] == pytest.approx(np.array([1.0, 1.0]),
                                                                         1e-3)
            assert ray_trace.image_plane.deflections.image[0] == pytest.approx(np.array([0.0, 0.0]), 1e-3)
            assert ray_trace.source_plane.grids.image[0] == pytest.approx(np.array([1.0, 1.0]),
                                                                          1e-3)

        def test__image_grid__sis_lens__image_coordinates_are_grid_and_source_plane_is_deflected(self,
                                                                                                 grids,
                                                                                                 galaxy_mass_sis):
            ray_trace = ray_tracing.Tracer(lens_galaxies=[galaxy_mass_sis], source_galaxies=galaxy_mass_sis,
                                           image_plane_grids=grids)

            assert ray_trace.image_plane.grids.image[0] == pytest.approx(np.array([1.0, 1.0]),
                                                                         1e-3)
            assert ray_trace.image_plane.deflections.image[0] == pytest.approx(np.array([0.707, 0.707]), 1e-3)
            assert ray_trace.source_plane.grids.image[0] == pytest.approx(
                np.array([1.0 - 0.707, 1.0 - 0.707]), 1e-3)

        def test__image_grid__2_sis_lenses__same_as_above_but_deflections_double(self, grids,
                                                                                 galaxy_mass_sis):
            ray_trace = ray_tracing.Tracer(lens_galaxies=[galaxy_mass_sis, galaxy_mass_sis],
                                           source_galaxies=galaxy_mass_sis,
                                           image_plane_grids=grids)

            assert ray_trace.image_plane.grids.image[0] == pytest.approx(np.array([1.0, 1.0]),
                                                                         1e-3)
            assert ray_trace.image_plane.deflections.image[0] == pytest.approx(np.array([1.414, 1.414]), 1e-3)
            assert ray_trace.source_plane.grids.image[0] == pytest.approx(
                np.array([1.0 - 1.414, 1.0 - 1.414]), 1e-3)

        def test__grids__sis_lens__planes_setup_correctly(self, grids, galaxy_mass_sis):
            ray_trace = ray_tracing.Tracer(lens_galaxies=[galaxy_mass_sis], source_galaxies=galaxy_mass_sis,
                                           image_plane_grids=grids)

            assert ray_trace.image_plane.grids.image[0] == pytest.approx(np.array([1.0, 1.0]),
                                                                         1e-3)
            assert ray_trace.image_plane.grids.sub[0] == pytest.approx(
                np.array([1.0, 1.0]), 1e-3)
            assert ray_trace.image_plane.grids.sub[1] == pytest.approx(
                np.array([1.0, 0.0]), 1e-3)
            assert ray_trace.image_plane.grids.sub[2] == pytest.approx(
                np.array([1.0, 1.0]), 1e-3)
            assert ray_trace.image_plane.grids.sub[3] == pytest.approx(
                np.array([1.0, 0.0]), 1e-3)
            assert ray_trace.image_plane.grids.blurring[0] == pytest.approx(
                np.array([1.0, 0.0]), 1e-3)

            assert ray_trace.image_plane.deflections.image[0] == pytest.approx(np.array([0.707, 0.707]), 1e-3)
            assert ray_trace.image_plane.deflections.sub[0] == pytest.approx(np.array([0.707, 0.707]), 1e-3)
            assert ray_trace.image_plane.deflections.sub[1] == pytest.approx(np.array([1.0, 0.0]), 1e-3)
            assert ray_trace.image_plane.deflections.sub[2] == pytest.approx(np.array([0.707, 0.707]), 1e-3)
            assert ray_trace.image_plane.deflections.sub[3] == pytest.approx(np.array([1.0, 0.0]), 1e-3)
            assert ray_trace.image_plane.deflections.blurring[0] == pytest.approx(np.array([1.0, 0.0]), 1e-3)

            assert ray_trace.source_plane.grids.image[0] == pytest.approx(
                np.array([1.0 - 0.707, 1.0 - 0.707]),
                1e-3)
            assert ray_trace.source_plane.grids.sub[0] == pytest.approx(
                np.array([1.0 - 0.707, 1.0 - 0.707]), 1e-3)
            assert ray_trace.source_plane.grids.sub[1] == pytest.approx(
                np.array([0.0, 0.0]), 1e-3)
            assert ray_trace.source_plane.grids.sub[2] == pytest.approx(
                np.array([1.0 - 0.707, 1.0 - 0.707]), 1e-3)
            assert ray_trace.source_plane.grids.sub[3] == pytest.approx(
                np.array([0.0, 0.0]), 1e-3)
            assert ray_trace.source_plane.grids.blurring[0] == pytest.approx(
                np.array([0.0, 0.0]), 1e-3)

    class TestImageFromGalaxies:

        def test__no_galaxies__image_is_sum_of_image_plane_and_source_plane_images(self, grids,
                                                                                   no_galaxies):
            image_plane = ray_tracing.Plane(galaxies=no_galaxies, grids=grids,
                                            compute_deflections=True)
            source_plane = ray_tracing.Plane(galaxies=no_galaxies, grids=grids,
                                             compute_deflections=False)
            plane_image = image_plane.generate_image_of_galaxy_light_profiles(
            ) + source_plane.generate_image_of_galaxy_light_profiles()

            ray_trace = ray_tracing.Tracer(lens_galaxies=no_galaxies, source_galaxies=no_galaxies,
                                           image_plane_grids=grids)
            ray_trace_image = ray_trace.generate_image_of_galaxy_light_profiles()

            assert (plane_image == ray_trace_image).all()

        def test__galaxy_light_sersic_no_mass_image_is_sum_of_image_plane_and_source_plane(self,
                                                                                           light_only_source_plane,
                                                                                           light_only_image_plane,
                                                                                           light_only_ray_tracer):
            plane_image = light_only_image_plane.generate_image_of_galaxy_light_profiles(
            ) + light_only_source_plane.generate_image_of_galaxy_light_profiles()

            ray_trace_image = light_only_ray_tracer.generate_image_of_galaxy_light_profiles()

            assert (plane_image == ray_trace_image).all()

        def test__plane_galaxy_images(self, light_only_source_plane, light_only_image_plane):
            assert (light_only_image_plane.galaxy_images[0] ==
                    light_only_image_plane.generate_image_of_galaxy_light_profiles()).all()
            assert (light_only_source_plane.galaxy_images[0] ==
                    light_only_source_plane.generate_image_of_galaxy_light_profiles()).all()

        def test__tracer_galaxy_images(self, light_only_ray_tracer):
            galaxy_images = light_only_ray_tracer.galaxy_images
            assert (np.add(galaxy_images[0],
                           galaxy_images[1]) == light_only_ray_tracer.generate_image_of_galaxy_light_profiles()).all()

        def test__galaxy_light_sersic_mass_sis__source_plane_image_includes_deflections(self, grids,
                                                                                        galaxy_light_and_mass):
            image_plane = ray_tracing.Plane(galaxies=[galaxy_light_and_mass],
                                            grids=grids,
                                            compute_deflections=True)
            deflections_grid = ray_tracing.deflections_for_grids(grids,
                                                                 galaxies=[galaxy_light_and_mass])
            source_grid = ray_tracing.traced_collection_for_deflections(grids, deflections_grid)
            source_plane = ray_tracing.Plane(galaxies=[galaxy_light_and_mass], grids=source_grid,
                                             compute_deflections=False)
            plane_image = image_plane.generate_image_of_galaxy_light_profiles(
            ) + source_plane.generate_image_of_galaxy_light_profiles()

            ray_trace = ray_tracing.Tracer(lens_galaxies=[galaxy_light_and_mass],
                                           source_galaxies=[galaxy_light_and_mass],
                                           image_plane_grids=grids)
            ray_trace_image = ray_trace.generate_image_of_galaxy_light_profiles()

            assert (plane_image == ray_trace_image).all()

    class TestBlurringImageFromGalaxies:

        def test__no_galaxies__image_is_sum_of_image_plane_and_source_plane_images(self, grids,
                                                                                   no_galaxies):
            image_plane = ray_tracing.Plane(galaxies=no_galaxies, grids=grids,
                                            compute_deflections=True)
            source_plane = ray_tracing.Plane(galaxies=no_galaxies, grids=grids,
                                             compute_deflections=False)
            plane_image = image_plane.blurring_image_from_galaxy_light_profiles(

            ) + source_plane.blurring_image_from_galaxy_light_profiles()

            ray_trace = ray_tracing.Tracer(lens_galaxies=no_galaxies, source_galaxies=no_galaxies,
                                           image_plane_grids=grids)
            ray_trace_image = ray_trace.generate_blurring_image_of_galaxy_light_profiles()

            assert (plane_image == ray_trace_image).all()

        def test__galaxy_light_sersic_no_mass_image_is_sum_of_image_plane_and_source_plane_images(self,
                                                                                                  grids,
                                                                                                  galaxy_light_only):
            image_plane = ray_tracing.Plane(galaxies=[galaxy_light_only], grids=grids,
                                            compute_deflections=True)
            source_plane = ray_tracing.Plane(galaxies=[galaxy_light_only], grids=grids,
                                             compute_deflections=False)
            plane_image = image_plane.blurring_image_from_galaxy_light_profiles(
            ) + source_plane.blurring_image_from_galaxy_light_profiles()

            ray_trace = ray_tracing.Tracer(lens_galaxies=[galaxy_light_only],
                                           source_galaxies=[galaxy_light_only],
                                           image_plane_grids=grids)
            ray_trace_image = ray_trace.generate_blurring_image_of_galaxy_light_profiles()

            assert (plane_image == ray_trace_image).all()

        def test__galaxy_light_sersic_mass_sis__source_plane_image_includes_deflections(self, grids,
                                                                                        galaxy_light_and_mass):
            image_plane = ray_tracing.Plane(galaxies=[galaxy_light_and_mass],
                                            grids=grids,
                                            compute_deflections=True)
            deflections_grid = ray_tracing.deflections_for_grids(grids,
                                                                 galaxies=[galaxy_light_and_mass])
            source_grid = ray_tracing.traced_collection_for_deflections(grids, deflections_grid)
            source_plane = ray_tracing.Plane(galaxies=[galaxy_light_and_mass], grids=source_grid,
                                             compute_deflections=False)
            plane_image = image_plane.blurring_image_from_galaxy_light_profiles(
            ) + source_plane.blurring_image_from_galaxy_light_profiles()

            ray_trace = ray_tracing.Tracer(lens_galaxies=[galaxy_light_and_mass],
                                           source_galaxies=[galaxy_light_and_mass],
                                           image_plane_grids=grids)
            ray_trace_image = ray_trace.generate_blurring_image_of_galaxy_light_profiles()

            assert (plane_image == ray_trace_image).all()

    class TestReconstructorFromGalaxy:

        def test__no_galaxies_in_plane__returns_none(self, grids, sparse_mask):
            galaxy_no_pix = galaxy.Galaxy()

            tracing = ray_tracing.Tracer(lens_galaxies=[], source_galaxies=[galaxy_no_pix],
                                         image_plane_grids=grids)

            reconstructors = tracing.reconstructors_from_source_plane(MockBorders(), sparse_mask)

            assert reconstructors is None

        def test__image_galaxy_has_pixelization__still_returns_none(self, grids, sparse_mask):
            galaxy_pix = galaxy.Galaxy(pixelization=MockPixelization(value=1))
            galaxy_no_pix = galaxy.Galaxy()

            tracing = ray_tracing.Tracer(lens_galaxies=[galaxy_pix], source_galaxies=[galaxy_no_pix],
                                         image_plane_grids=grids)

            reconstructors = tracing.reconstructors_from_source_plane(MockBorders(), sparse_mask)

            assert reconstructors is None

        def test__source_galaxy_has_pixelization__returns_reconstructor(self, grids, sparse_mask):
            galaxy_pix = galaxy.Galaxy(pixelization=MockPixelization(value=1))
            galaxy_no_pix = galaxy.Galaxy()

            tracing = ray_tracing.Tracer(lens_galaxies=[galaxy_no_pix], source_galaxies=[galaxy_pix],
                                         image_plane_grids=grids)

            reconstructors = tracing.reconstructors_from_source_plane(MockBorders(), sparse_mask)

            assert reconstructors == 1


class TestMultiTracer(object):
    class TestGalaxyOrder:

        def test__3_galaxies_reordered_in_ascending_redshift(self, grids):
            tracer = ray_tracing.MultiTracer(galaxies=[galaxy.Galaxy(redshift=2.0), galaxy.Galaxy(redshift=1.0),
                                                       galaxy.Galaxy(redshift=0.1)],
                                             image_plane_grids=grids,
                                             cosmology=cosmo.Planck15)

            assert tracer.galaxies_redshift_order[0].redshift == 0.1
            assert tracer.galaxies_redshift_order[1].redshift == 1.0
            assert tracer.galaxies_redshift_order[2].redshift == 2.0

        def test_3_galaxies_two_same_redshift_planes_redshift_order_is_size_2_with_redshifts(self,
                                                                                             grids):
            tracer = ray_tracing.MultiTracer(galaxies=[galaxy.Galaxy(redshift=1.0), galaxy.Galaxy(redshift=1.0),
                                                       galaxy.Galaxy(redshift=0.1)],
                                             image_plane_grids=grids,
                                             cosmology=cosmo.Planck15)

            assert tracer.galaxies_redshift_order[0].redshift == 0.1
            assert tracer.galaxies_redshift_order[1].redshift == 1.0
            assert tracer.galaxies_redshift_order[2].redshift == 1.0

            assert tracer.planes_redshift_order[0] == 0.1
            assert tracer.planes_redshift_order[1] == 1.0

        def test__6_galaxies_producing_4_planes(self, grids):
            g0 = galaxy.Galaxy(redshift=1.0)
            g1 = galaxy.Galaxy(redshift=1.0)
            g2 = galaxy.Galaxy(redshift=0.1)
            g3 = galaxy.Galaxy(redshift=1.05)
            g4 = galaxy.Galaxy(redshift=0.95)
            g5 = galaxy.Galaxy(redshift=1.05)

            tracer = ray_tracing.MultiTracer(galaxies=[g0, g1, g2, g3, g4, g5],
                                             image_plane_grids=grids, cosmology=cosmo.Planck15)

            assert tracer.galaxies_redshift_order[0].redshift == 0.1
            assert tracer.galaxies_redshift_order[1].redshift == 0.95
            assert tracer.galaxies_redshift_order[2].redshift == 1.0
            assert tracer.galaxies_redshift_order[3].redshift == 1.0
            assert tracer.galaxies_redshift_order[4].redshift == 1.05
            assert tracer.galaxies_redshift_order[5].redshift == 1.05

            assert tracer.planes_redshift_order[0] == 0.1
            assert tracer.planes_redshift_order[1] == 0.95
            assert tracer.planes_redshift_order[2] == 1.0
            assert tracer.planes_redshift_order[3] == 1.05

        def test__6_galaxies__plane_galaxies_are_correct(self, grids):
            g0 = galaxy.Galaxy(redshift=1.0)
            g1 = galaxy.Galaxy(redshift=1.0)
            g2 = galaxy.Galaxy(redshift=0.1)
            g3 = galaxy.Galaxy(redshift=1.05)
            g4 = galaxy.Galaxy(redshift=0.95)
            g5 = galaxy.Galaxy(redshift=1.05)

            tracer = ray_tracing.MultiTracer(galaxies=[g0, g1, g2, g3, g4, g5],
                                             image_plane_grids=grids, cosmology=cosmo.Planck15)

            assert tracer.planes_galaxies[0] == [g2]
            assert tracer.planes_galaxies[1] == [g4]
            assert tracer.planes_galaxies[2] == [g0, g1]
            assert tracer.planes_galaxies[3] == [g3, g5]

    class TestRayTracingPlanes:

        def test__6_galaxies__tracer_planes_are_correct(self, grids):
            g0 = galaxy.Galaxy(redshift=2.0)
            g1 = galaxy.Galaxy(redshift=2.0)
            g2 = galaxy.Galaxy(redshift=0.1)
            g3 = galaxy.Galaxy(redshift=3.0)
            g4 = galaxy.Galaxy(redshift=1.0)
            g5 = galaxy.Galaxy(redshift=3.0)

            tracer = ray_tracing.MultiTracer(galaxies=[g0, g1, g2, g3, g4, g5],
                                             image_plane_grids=grids, cosmology=cosmo.Planck15)

            assert tracer.planes[0].galaxies == [g2]
            assert tracer.planes[1].galaxies == [g4]
            assert tracer.planes[2].galaxies == [g0, g1]
            assert tracer.planes[3].galaxies == [g3, g5]

        def test__4_planes__coordinate_grids_and_deflections_are_correct__sis_mass_profile(self,
                                                                                           grids):
            import math

            g0 = galaxy.Galaxy(redshift=2.0, mass_profile=mass_profiles.SphericalIsothermal(einstein_radius=1.0))
            g1 = galaxy.Galaxy(redshift=2.0, mass_profile=mass_profiles.SphericalIsothermal(einstein_radius=1.0))
            g2 = galaxy.Galaxy(redshift=0.1, mass_profile=mass_profiles.SphericalIsothermal(einstein_radius=1.0))
            g3 = galaxy.Galaxy(redshift=3.0, mass_profile=mass_profiles.SphericalIsothermal(einstein_radius=1.0))
            g4 = galaxy.Galaxy(redshift=1.0, mass_profile=mass_profiles.SphericalIsothermal(einstein_radius=1.0))
            g5 = galaxy.Galaxy(redshift=3.0, mass_profile=mass_profiles.SphericalIsothermal(einstein_radius=1.0))

            tracer = ray_tracing.MultiTracer(galaxies=[g0, g1, g2, g3, g4, g5],
                                             image_plane_grids=grids, cosmology=cosmo.Planck15)

            # From unit test below:
            # Beta_01 = 0.9348
            # Beta_02 = 0.9840
            # Beta_03 = 1.0
            # Beta_12 = 0.754
            # Beta_13 = 1.0
            # Beta_23 = 1.0

            val = math.sqrt(2) / 2.0

            assert tracer.planes[0].grids.image[0] == pytest.approx(np.array([1.0, 1.0]), 1e-4)
            assert tracer.planes[0].grids.sub[0] == pytest.approx(np.array([1.0, 1.0]),
                                                                  1e-4)
            assert tracer.planes[0].grids.sub[1] == pytest.approx(np.array([1.0, 0.0]),
                                                                  1e-4)
            assert tracer.planes[0].grids.blurring[0] == pytest.approx(np.array([1.0, 0.0]),
                                                                       1e-4)
            assert tracer.planes[0].deflections.image[0] == pytest.approx(np.array([val, val]), 1e-4)
            assert tracer.planes[0].deflections.sub[0] == pytest.approx(np.array([val, val]), 1e-4)
            assert tracer.planes[0].deflections.sub[1] == pytest.approx(np.array([1.0, 0.0]), 1e-4)
            assert tracer.planes[0].deflections.blurring[0] == pytest.approx(np.array([1.0, 0.0]), 1e-4)

            assert tracer.planes[1].grids.image[0] == pytest.approx(
                np.array([(1.0 - 0.9348 * val), (1.0 - 0.9348 * val)]), 1e-4)
            assert tracer.planes[1].grids.sub[0] == pytest.approx(
                np.array([(1.0 - 0.9348 * val), (1.0 - 0.9348 * val)]), 1e-4)
            assert tracer.planes[1].grids.sub[1] == pytest.approx(
                np.array([(1.0 - 0.9348 * 1.0), 0.0]), 1e-4)
            assert tracer.planes[1].grids.blurring[0] == pytest.approx(
                np.array([(1.0 - 0.9348 * 1.0), 0.0]), 1e-4)

            defl11 = g0.deflections_from_grid(grid=np.array([[(1.0 - 0.9348 * val), (1.0 - 0.9348 * val)]]))
            defl12 = g0.deflections_from_grid(grid=np.array([[(1.0 - 0.9348 * 1.0), 0.0]]))

            assert tracer.planes[1].deflections.image[0] == pytest.approx(defl11[0], 1e-4)
            assert tracer.planes[1].deflections.sub[0] == pytest.approx(defl11[0], 1e-4)
            assert tracer.planes[1].deflections.sub[1] == pytest.approx(defl12[0], 1e-4)
            assert tracer.planes[1].deflections.blurring[0] == pytest.approx(defl12[0], 1e-4)

            assert tracer.planes[2].grids.image[0] == pytest.approx(
                np.array([(1.0 - 0.9839601 * val - 0.7539734 * defl11[0, 0]),
                          (1.0 - 0.9839601 * val - 0.7539734 * defl11[0, 1])]), 1e-4)
            assert tracer.planes[2].grids.sub[0] == pytest.approx(
                np.array([(1.0 - 0.9839601 * val - 0.7539734 * defl11[0, 0]),
                          (1.0 - 0.9839601 * val - 0.7539734 * defl11[0, 1])]), 1e-4)
            assert tracer.planes[2].grids.sub[1] == pytest.approx(
                np.array([(1.0 - 0.9839601 * 1.0 - 0.7539734 * defl12[0, 0]),
                          0.0]), 1e-4)
            assert tracer.planes[2].grids.blurring[0] == pytest.approx(
                np.array([(1.0 - 0.9839601 * 1.0 - 0.7539734 * defl12[0, 0]),
                          0.0]), 1e-4)

            # 2 Galaxies in this plane, so multiply by 2.0

            defl21 = 2.0 * g0.deflections_from_grid(
                grid=np.array([[(1.0 - 0.9839601 * val - 0.7539734 * defl11[0, 0]),
                                (1.0 - 0.9839601 * val - 0.7539734 * defl11[0, 1])]]))
            defl22 = 2.0 * g0.deflections_from_grid(
                grid=np.array([[(1.0 - 0.9839601 * 1.0 - 0.7539734 * defl12[0, 0]),
                                0.0]]))

            assert tracer.planes[2].deflections.image[0] == pytest.approx(defl21[0], 1e-4)
            assert tracer.planes[2].deflections.sub[0] == pytest.approx(defl21[0], 1e-4)
            assert tracer.planes[2].deflections.sub[1] == pytest.approx(defl22[0], 1e-4)
            assert tracer.planes[2].deflections.blurring[0] == pytest.approx(defl22[0], 1e-4)

            coord1 = (1.0 - tracer.planes[0].deflections.image[0, 0] - tracer.planes[1].deflections.image[
                0, 0] -
                      tracer.planes[2].deflections.image[0, 0])

            coord2 = (1.0 - tracer.planes[0].deflections.image[0, 1] - tracer.planes[1].deflections.image[
                0, 1] -
                      tracer.planes[2].deflections.image[0, 1])

            coord3 = (1.0 - tracer.planes[0].deflections.sub[1, 0] -
                      tracer.planes[1].deflections.sub[1, 0] -
                      tracer.planes[2].deflections.sub[1, 0])

            assert tracer.planes[3].grids.image[0] == pytest.approx(np.array([coord1, coord2]),
                                                                    1e-4)
            assert tracer.planes[3].grids.sub[0] == pytest.approx(
                np.array([coord1, coord2]), 1e-4)
            assert tracer.planes[3].grids.sub[1] == pytest.approx(np.array([coord3, 0.0]),
                                                                  1e-4)
            assert tracer.planes[3].grids.blurring[0] == pytest.approx(np.array([coord3, 0.0]),
                                                                       1e-4)

    class TestImageFromGalaxies:

        def test__galaxy_light_sersic_no_mass__image_sum_of_all_3_planes(self, grids):
            sersic = light_profiles.EllipticalSersic(axis_ratio=0.5, phi=0.0, intensity=1.0, effective_radius=0.6,
                                                     sersic_index=4.0)

            g0 = galaxy.Galaxy(redshift=0.1, light_profile=sersic)
            g1 = galaxy.Galaxy(redshift=1.0, light_profile=sersic)
            g2 = galaxy.Galaxy(redshift=2.0, light_profile=sersic)

            ray_trace = ray_tracing.MultiTracer(galaxies=[g0, g1, g2], image_plane_grids=grids,
                                                cosmology=cosmo.Planck15)
            ray_trace_image = ray_trace.generate_image_of_galaxy_light_profiles()

            plane_0 = ray_tracing.Plane(galaxies=[g0], grids=grids,
                                        compute_deflections=True)
            plane_1 = ray_tracing.Plane(galaxies=[g1], grids=grids,
                                        compute_deflections=True)
            plane_2 = ray_tracing.Plane(galaxies=[g2], grids=grids,
                                        compute_deflections=False)

            plane_image = (plane_0.generate_image_of_galaxy_light_profiles() +
                           plane_1.generate_image_of_galaxy_light_profiles() +
                           plane_2.generate_image_of_galaxy_light_profiles())

            assert (plane_image == ray_trace_image).all()

        def test__galaxy_light_sersic_mass_sis__source_plane_image_includes_deflections(self, grids):
            sersic = light_profiles.EllipticalSersic(axis_ratio=0.5, phi=0.0, intensity=1.0, effective_radius=0.6,
                                                     sersic_index=4.0)

            sis = mass_profiles.SphericalIsothermal(einstein_radius=1.0)

            g0 = galaxy.Galaxy(redshift=0.1, light_profile=sersic, mass_profile=sis)
            g1 = galaxy.Galaxy(redshift=1.0, light_profile=sersic, mass_profile=sis)
            g2 = galaxy.Galaxy(redshift=2.0, light_profile=sersic, mass_profile=sis)

            ray_trace = ray_tracing.MultiTracer(galaxies=[g0, g1, g2], image_plane_grids=grids,
                                                cosmology=cosmo.Planck15)
            ray_trace_image = ray_trace.generate_image_of_galaxy_light_profiles()

            plane_0 = ray_trace.planes[0]
            plane_1 = ray_trace.planes[1]
            plane_2 = ray_trace.planes[2]

            plane_image = (plane_0.generate_image_of_galaxy_light_profiles() +
                           plane_1.generate_image_of_galaxy_light_profiles() +
                           plane_2.generate_image_of_galaxy_light_profiles())

            assert (plane_image == ray_trace_image).all()

    class TestBlurringImageFromGalaxies:

        def test__galaxy_light_sersic_no_mass__image_sum_of_all_3_planes(self, grids):
            sersic = light_profiles.EllipticalSersic(axis_ratio=0.5, phi=0.0, intensity=1.0, effective_radius=0.6,
                                                     sersic_index=4.0)

            g0 = galaxy.Galaxy(redshift=0.1, light_profile=sersic)
            g1 = galaxy.Galaxy(redshift=1.0, light_profile=sersic)
            g2 = galaxy.Galaxy(redshift=2.0, light_profile=sersic)

            ray_trace = ray_tracing.MultiTracer(galaxies=[g0, g1, g2], image_plane_grids=grids,
                                                cosmology=cosmo.Planck15)
            ray_trace_image = ray_trace.generate_blurring_image_of_galaxy_light_profiles()

            plane_0 = ray_tracing.Plane(galaxies=[g0], grids=grids,
                                        compute_deflections=True)
            plane_1 = ray_tracing.Plane(galaxies=[g1], grids=grids,
                                        compute_deflections=True)
            plane_2 = ray_tracing.Plane(galaxies=[g2], grids=grids,
                                        compute_deflections=False)

            plane_image = (plane_0.blurring_image_from_galaxy_light_profiles() +
                           plane_1.blurring_image_from_galaxy_light_profiles() +
                           plane_2.blurring_image_from_galaxy_light_profiles())

            assert (plane_image == ray_trace_image).all()

        def test__galaxy_light_sersic_mass_sis__source_plane_image_includes_deflections(self, grids):
            sersic = light_profiles.EllipticalSersic(axis_ratio=0.5, phi=0.0, intensity=1.0, effective_radius=0.6,
                                                     sersic_index=4.0)

            sis = mass_profiles.SphericalIsothermal(einstein_radius=1.0)

            g0 = galaxy.Galaxy(redshift=0.1, light_profile=sersic, mass_profile=sis)
            g1 = galaxy.Galaxy(redshift=1.0, light_profile=sersic, mass_profile=sis)
            g2 = galaxy.Galaxy(redshift=2.0, light_profile=sersic, mass_profile=sis)

            ray_trace = ray_tracing.MultiTracer(galaxies=[g0, g1, g2], image_plane_grids=grids,
                                                cosmology=cosmo.Planck15)
            ray_trace_image = ray_trace.generate_blurring_image_of_galaxy_light_profiles()

            plane_0 = ray_trace.planes[0]
            plane_1 = ray_trace.planes[1]
            plane_2 = ray_trace.planes[2]

            plane_image = (plane_0.blurring_image_from_galaxy_light_profiles() +
                           plane_1.blurring_image_from_galaxy_light_profiles() +
                           plane_2.blurring_image_from_galaxy_light_profiles())

            assert (plane_image == ray_trace_image).all()

    class TestReconstructorFromGalaxy:

        def test__3_galaxies__non_have_pixelization__returns_none_x3(self, grids, sparse_mask):
            sis = mass_profiles.SphericalIsothermal(einstein_radius=1.0)

            g0 = galaxy.Galaxy(redshift=0.1, mass_profile=sis)
            g1 = galaxy.Galaxy(redshift=1.0, mass_profile=sis)
            g2 = galaxy.Galaxy(redshift=2.0, mass_profile=sis)

            tracing = ray_tracing.MultiTracer(galaxies=[g0, g1, g2], image_plane_grids=grids,
                                              cosmology=cosmo.Planck15)

            reconstructors = tracing.reconstructors_from_planes(MockBorders(), sparse_mask)

            assert reconstructors == [None, None, None]

        def test__3_galaxies__1_has_pixelization__returns_none_x2_and_pixelization(self, grids,
                                                                                   sparse_mask):
            sis = mass_profiles.SphericalIsothermal(einstein_radius=1.0)

            g0 = galaxy.Galaxy(redshift=0.1, mass_profile=sis)
            g1 = galaxy.Galaxy(redshift=1.0, pixelization=MockPixelization(value=1), mass_profile=sis)
            g2 = galaxy.Galaxy(redshift=2.0, mass_profile=sis)

            tracing = ray_tracing.MultiTracer(galaxies=[g0, g1, g2], image_plane_grids=grids,
                                              cosmology=cosmo.Planck15)

            reconstructors = tracing.reconstructors_from_planes(MockBorders(), sparse_mask)

            assert reconstructors == [None, 1, None]

        def test__3_galaxies__all_have_pixelization__returns_pixelizations(self, grids, sparse_mask):
            sis = mass_profiles.SphericalIsothermal(einstein_radius=1.0)

            g0 = galaxy.Galaxy(redshift=0.1, pixelization=MockPixelization(value=0.5), mass_profile=sis)
            g1 = galaxy.Galaxy(redshift=1.0, pixelization=MockPixelization(value=1), mass_profile=sis)
            g2 = galaxy.Galaxy(redshift=2.0, pixelization=MockPixelization(value=2), mass_profile=sis)

            tracing = ray_tracing.MultiTracer(galaxies=[g0, g1, g2], image_plane_grids=grids,
                                              cosmology=cosmo.Planck15)

            reconstructors = tracing.reconstructors_from_planes(MockBorders(), sparse_mask)

            assert reconstructors == [0.5, 1, 2]


class TestPlane(object):
    class TestBasicSetup:

        def test__collection__sis_lens__coordinates_and_deflections_setup_for_every_grid(self,
                                                                                         grids,
                                                                                         galaxy_mass_sis):
            plane = ray_tracing.Plane(galaxies=[galaxy_mass_sis], grids=grids)

            assert plane.grids.image[0] == pytest.approx(np.array([1.0, 1.0]), 1e-3)
            assert plane.grids.sub[0] == pytest.approx(np.array([1.0, 1.0]), 1e-3)
            assert plane.grids.sub[1] == pytest.approx(np.array([1.0, 0.0]), 1e-3)
            assert plane.grids.sub[2] == pytest.approx(np.array([1.0, 1.0]), 1e-3)
            assert plane.grids.sub[3] == pytest.approx(np.array([1.0, 0.0]), 1e-3)
            assert plane.grids.blurring[0] == pytest.approx(np.array([1.0, 0.0]), 1e-3)

        def test__collection__3_identical_sis_lenses__deflections_triple_compared_to_above(self,
                                                                                           grids,
                                                                                           galaxy_mass_sis):
            lens_plane = ray_tracing.Plane(galaxies=[galaxy_mass_sis, galaxy_mass_sis, galaxy_mass_sis],
                                           grids=grids, compute_deflections=True)

            assert lens_plane.grids.image[0] == pytest.approx(np.array([1.0, 1.0]), 1e-5)
            assert lens_plane.grids.sub[0] == pytest.approx(np.array([1.0, 1.0]), 1e-5)
            assert lens_plane.grids.sub[1] == pytest.approx(np.array([1.0, 0.0]), 1e-5)
            assert lens_plane.grids.sub[2] == pytest.approx(np.array([1.0, 1.0]), 1e-5)
            assert lens_plane.grids.sub[3] == pytest.approx(np.array([1.0, 0.0]), 1e-5)
            assert lens_plane.grids.blurring[0] == pytest.approx(np.array([1.0, 0.0]), 1e-5)

            assert lens_plane.deflections.image[0] == pytest.approx(np.array([3.0 * 0.707, 3.0 * 0.707]), 1e-3)
            assert lens_plane.deflections.sub[0] == pytest.approx(np.array([3.0 * 0.707, 3.0 * 0.707]),
                                                                  1e-3)
            assert lens_plane.deflections.sub[1] == pytest.approx(np.array([3.0, 0.0]), 1e-3)
            assert lens_plane.deflections.sub[2] == pytest.approx(np.array([3.0 * 0.707, 3.0 * 0.707]),
                                                                  1e-3)
            assert lens_plane.deflections.sub[3] == pytest.approx(np.array([3.0, 0.0]), 1e-3)
            assert lens_plane.deflections.blurring[0] == pytest.approx(np.array([3.0, 0.0]), 1e-3)

        def test__collection__lens_is_3_identical_sis_profiles__deflections_triple_like_above(self,
                                                                                              grids,
                                                                                              lens_sis_x3):
            lens_plane = ray_tracing.Plane(galaxies=[lens_sis_x3], grids=grids,
                                           compute_deflections=True)

            assert lens_plane.grids.image[0] == pytest.approx(np.array([1.0, 1.0]), 1e-5)
            assert lens_plane.grids.sub[0] == pytest.approx(np.array([1.0, 1.0]), 1e-5)
            assert lens_plane.grids.sub[1] == pytest.approx(np.array([1.0, 0.0]), 1e-5)
            assert lens_plane.grids.sub[2] == pytest.approx(np.array([1.0, 1.0]), 1e-5)
            assert lens_plane.grids.sub[3] == pytest.approx(np.array([1.0, 0.0]), 1e-5)
            assert lens_plane.grids.blurring[0] == pytest.approx(np.array([1.0, 0.0]), 1e-5)

            assert lens_plane.deflections.image[0] == pytest.approx(np.array([3.0 * 0.707, 3.0 * 0.707]), 1e-3)
            assert lens_plane.deflections.sub[0] == pytest.approx(np.array([3.0 * 0.707, 3.0 * 0.707]),
                                                                  1e-3)
            assert lens_plane.deflections.sub[1] == pytest.approx(np.array([3.0, 0.0]), 1e-3)
            assert lens_plane.deflections.sub[2] == pytest.approx(np.array([3.0 * 0.707, 3.0 * 0.707]),
                                                                  1e-3)
            assert lens_plane.deflections.sub[3] == pytest.approx(np.array([3.0, 0.0]), 1e-3)
            assert lens_plane.deflections.blurring[0] == pytest.approx(np.array([3.0, 0.0]), 1e-3)

        def test__grids__complex_mass_model(self, grids):
            power_law = mass_profiles.EllipticalPowerLaw(centre=(1.0, 4.0), axis_ratio=0.7, phi=30.0,
                                                         einstein_radius=1.0, slope=2.2)

            nfw = mass_profiles.SphericalNFW(kappa_s=0.1, scale_radius=5.0)

            lens_galaxy = galaxy.Galaxy(redshift=0.1, mass_profile_1=power_law, mass_profile_2=nfw)

            lens_plane = ray_tracing.Plane(galaxies=[lens_galaxy], grids=grids,
                                           compute_deflections=True)

            assert lens_plane.grids.image[0] == pytest.approx(np.array([1.0, 1.0]), 1e-5)
            assert lens_plane.grids.sub[0] == pytest.approx(np.array([1.0, 1.0]), 1e-5)
            assert lens_plane.grids.sub[1] == pytest.approx(np.array([1.0, 0.0]), 1e-5)
            assert lens_plane.grids.sub[2] == pytest.approx(np.array([1.0, 1.0]), 1e-5)
            assert lens_plane.grids.sub[3] == pytest.approx(np.array([1.0, 0.0]), 1e-5)
            assert lens_plane.grids.blurring[0] == pytest.approx(np.array([1.0, 0.0]), 1e-5)

    class TestImageFromGalaxies:

        def test__sersic_light_profile__intensities_equal_to_profile_and_galaxy_values(self, grids,
                                                                                       galaxy_light_sersic):
            sersic = galaxy_light_sersic.light_profiles[0]

            profile_intensity = sersic.intensity_from_grid(grids.sub)

            profile_intensity = (profile_intensity[0] + profile_intensity[1] + profile_intensity[2] +
                                 profile_intensity[3]) / 4

            galaxy_intensity = ray_tracing.intensities_via_sub_grid(grids.sub,
                                                                    galaxies=[galaxy_light_sersic])

            plane = ray_tracing.Plane(galaxies=[galaxy_light_sersic], grids=grids)
            image = plane.generate_image_of_galaxy_light_profiles()

            assert (image[0] == profile_intensity == galaxy_intensity).all()

        def test__same_as_above__now_with_multiple_sets_of_coordinates(self, grids,
                                                                       galaxy_light_sersic):
            sersic = galaxy_light_sersic.light_profiles[0]

            profile_intensity = sersic.intensity_from_grid(grids.sub)

            profile_intensity_image_0 = (profile_intensity[0] + profile_intensity[1] + profile_intensity[2] +
                                         profile_intensity[3]) / 4

            profile_intensity_image_1 = (profile_intensity[4] + profile_intensity[5] + profile_intensity[6] +
                                         profile_intensity[7]) / 4

            galaxy_intensity = ray_tracing.intensities_via_sub_grid(grids.sub,
                                                                    galaxies=[galaxy_light_sersic],
                                                                    )

            plane = ray_tracing.Plane(galaxies=[galaxy_light_sersic], grids=grids)
            image = plane.generate_image_of_galaxy_light_profiles()

            assert (image[0] == profile_intensity_image_0 == galaxy_intensity[0]).all()
            assert (image[1] == profile_intensity_image_1 == galaxy_intensity[1]).all()

        def test__same_as_above__now_galaxy_entered_3_times__intensities_triple(self, grids,
                                                                                galaxy_light_sersic):
            sersic = galaxy_light_sersic.light_profiles[0]

            profile_intensity = 3.0 * sersic.intensity_from_grid(grids.sub)

            profile_intensity_image_0 = (profile_intensity[0] + profile_intensity[1] + profile_intensity[2] +
                                         profile_intensity[3]) / 4

            profile_intensity_image_1 = (profile_intensity[4] + profile_intensity[5] + profile_intensity[6] +
                                         profile_intensity[7]) / 4

            galaxy_intensity = ray_tracing.intensities_via_sub_grid(grids.sub,
                                                                    galaxies=[galaxy_light_sersic, galaxy_light_sersic,
                                                                              galaxy_light_sersic])

            plane = ray_tracing.Plane(galaxies=[galaxy_light_sersic, galaxy_light_sersic, galaxy_light_sersic],
                                      grids=grids)

            image = plane.generate_image_of_galaxy_light_profiles()

            assert (image[0] == profile_intensity_image_0 == galaxy_intensity[0]).all()
            assert (image[1] == profile_intensity_image_1 == galaxy_intensity[1]).all()

    class TestBlurringImageFromGalaxies:

        def test__sersic_light_profile__intensities_equal_to_profile_and_galaxy_values(self, grids,
                                                                                       galaxy_light_sersic):
            sersic = galaxy_light_sersic.light_profiles[0]
            profile_intensity = sersic.intensity_from_grid(grids.blurring)

            blurring_galaxy_intensity = ray_tracing.intensities_via_grid(grids.blurring,
                                                                         galaxies=[galaxy_light_sersic])

            plane = ray_tracing.Plane(galaxies=[galaxy_light_sersic], grids=grids)
            blurring_image = plane.blurring_image_from_galaxy_light_profiles()

            assert (blurring_image[0] == profile_intensity[0] == blurring_galaxy_intensity).all()

        def test__same_as_above__now_with_multiple_sets_of_coordinates(self, grids,
                                                                       galaxy_light_sersic):
            grids.blurring = mask.ImageGrid(
                np.array([[1.0, 1.0], [5.0, 5.0], [-2.0, -9.0], [5.0, 7.0]]))

            sersic = galaxy_light_sersic.light_profiles[0]
            profile_intensity = sersic.intensity_from_grid(grids.blurring)

            blurring_galaxy_intensity = ray_tracing.intensities_via_grid(grids.blurring,
                                                                         galaxies=[galaxy_light_sersic])

            plane = ray_tracing.Plane(galaxies=[galaxy_light_sersic], grids=grids)
            blurring_image = plane.blurring_image_from_galaxy_light_profiles()

            assert (blurring_image[0] == profile_intensity[0] == blurring_galaxy_intensity[0]).all()
            assert (blurring_image[1] == profile_intensity[1] == blurring_galaxy_intensity[1]).all()
            assert (blurring_image[2] == profile_intensity[2] == blurring_galaxy_intensity[2]).all()
            assert (blurring_image[3] == profile_intensity[3] == blurring_galaxy_intensity[3]).all()

        def test__same_as_above__now_galaxy_entered_3_times__intensities_triple(self, grids,
                                                                                galaxy_light_sersic):
            grids.blurring = mask.ImageGrid(
                np.array([[1.0, 1.0], [5.0, 5.0], [-2.0, -9.0], [5.0, 7.0]]))

            sersic = galaxy_light_sersic.light_profiles[0]

            profile_intensity = sersic.intensity_from_grid(grids.blurring)

            blurring_galaxy_intensity = ray_tracing.intensities_via_grid(grids.blurring,
                                                                         galaxies=[galaxy_light_sersic])

            plane = ray_tracing.Plane(galaxies=[galaxy_light_sersic, galaxy_light_sersic, galaxy_light_sersic],
                                      grids=grids)
            blurring_image = plane.blurring_image_from_galaxy_light_profiles()

            assert (blurring_image[0] == 3.0 * profile_intensity[0] == 3.0 * blurring_galaxy_intensity[0]).all()
            assert (blurring_image[1] == 3.0 * profile_intensity[1] == 3.0 * blurring_galaxy_intensity[1]).all()
            assert (blurring_image[2] == 3.0 * profile_intensity[2] == 3.0 * blurring_galaxy_intensity[2]).all()
            assert (blurring_image[3] == 3.0 * profile_intensity[3] == 3.0 * blurring_galaxy_intensity[3]).all()

    class TestReconstructorFromGalaxies:

        def test__no_galaxies_with_pixelizations_in_plane__returns_none(self, grids, sparse_mask):
            galaxy_no_pix = galaxy.Galaxy()

            plane = ray_tracing.Plane(galaxies=[galaxy_no_pix], grids=grids)

            reconstructors = plane.reconstructor_from_plane(MockBorders(), sparse_mask)

            assert reconstructors is None

        def test__1_galaxy_in_plane__it_has_pixelization__extracts_reconstructor(self, grids, sparse_mask):
            galaxy_pix = galaxy.Galaxy(pixelization=MockPixelization(value=1))

            plane = ray_tracing.Plane(galaxies=[galaxy_pix], grids=grids)

            reconstructors = plane.reconstructor_from_plane(MockBorders(), sparse_mask)

            assert reconstructors == 1

        def test__2_galaxies_in_plane__1_has_pixelization__extracts_reconstructor(self, grids, sparse_mask):
            galaxy_pix = galaxy.Galaxy(pixelization=MockPixelization(value=1))
            galaxy_no_pix = galaxy.Galaxy()

            plane = ray_tracing.Plane(galaxies=[galaxy_no_pix, galaxy_pix],
                                      grids=grids)

            reconstructors = plane.reconstructor_from_plane(MockBorders(), sparse_mask)

            assert reconstructors == 1

        def test__2_galaxies_in_plane__both_have_pixelization__raises_error(self, grids, sparse_mask):
            galaxy_pix_0 = galaxy.Galaxy(pixelization=MockPixelization(value=1))
            galaxy_pix_1 = galaxy.Galaxy(pixelization=MockPixelization(value=2))

            plane = ray_tracing.Plane(galaxies=[galaxy_pix_0, galaxy_pix_1],
                                      grids=grids)

            with pytest.raises(exc.PixelizationException):
                plane.reconstructor_from_plane(MockBorders(), sparse_mask)


class TestDeflectionAnglesViaGalaxy(object):

    def test_all_coordinates(self, grids, galaxy_mass_sis):
        deflections = ray_tracing.deflections_for_grids(grids, [galaxy_mass_sis])

        assert deflections.image[0] == pytest.approx(np.array([0.707, 0.707]), 1e-3)
        assert deflections.sub[0] == pytest.approx(np.array([0.707, 0.707]), 1e-3)
        assert deflections.sub[1] == pytest.approx(np.array([1.0, 0.0]), 1e-3)
        #    assert deflections.sub.sub_grid_size == 2
        assert deflections.blurring[0] == pytest.approx(np.array([1.0, 0.0]), 1e-3)

    def test_three_identical_lenses__deflection_angles_triple(self, grids, galaxy_mass_sis):
        deflections = ray_tracing.deflections_for_grids(grids,
                                                        [galaxy_mass_sis, galaxy_mass_sis,
                                                         galaxy_mass_sis])

        assert deflections.image[0] == pytest.approx(np.array([3.0 * 0.707, 3.0 * 0.707]), 1e-3)
        assert deflections.sub[0] == pytest.approx(np.array([3.0 * 0.707, 3.0 * 0.707]), 1e-3)
        assert deflections.sub[1] == pytest.approx(np.array([3.0, 0.0]), 1e-3)
        #    assert deflections.sub.sub_grid_size == 2
        assert deflections.blurring[0] == pytest.approx(np.array([3.0, 0.0]), 1e-3)

    def test_one_lens_with_three_identical_mass_profiles__deflection_angles_triple(self, grids,
                                                                                   lens_sis_x3):
        deflections = ray_tracing.deflections_for_grids(grids, [lens_sis_x3])

        assert deflections.image[0] == pytest.approx(np.array([3.0 * 0.707, 3.0 * 0.707]), 1e-3)
        assert deflections.sub[0] == pytest.approx(np.array([3.0 * 0.707, 3.0 * 0.707]), 1e-3)
        assert deflections.sub[1] == pytest.approx(np.array([3.0, 0.0]), 1e-3)
        #    assert deflections.sub.sub_grid_size == 2
        assert deflections.blurring[0] == pytest.approx(np.array([3.0, 0.0]), 1e-3)


class TestIntensityViaGrid:

    def test__no_galaxies__intensities_returned_as_0s(self, grids, galaxy_no_profiles):
        grids.image = np.array([[1.0, 1.0],
                                [2.0, 2.0],
                                [3.0, 3.0]])

        intensities = ray_tracing.intensities_via_grid(image_grid=grids.image,
                                                       galaxies=[galaxy_no_profiles])

        assert (intensities[0] == np.array([0.0, 0.0])).all()
        assert (intensities[1] == np.array([0.0, 0.0])).all()
        assert (intensities[2] == np.array([0.0, 0.0])).all()

    def test__galaxy_sersic_light__intensities_returned_as_correct_values(self, grids,
                                                                          galaxy_light_sersic):
        grids.image = np.array([[1.0, 1.0],
                                [1.0, 0.0],
                                [-1.0, 0.0]])

        galaxy_intensities = galaxy_light_sersic.intensity_from_grid(grids.image)

        tracer_intensities = ray_tracing.intensities_via_grid(image_grid=grids.image,
                                                              galaxies=[galaxy_light_sersic])

        assert (galaxy_intensities == tracer_intensities).all()

    def test__galaxy_sis_mass_x3__intensities_tripled_from_above(self, grids,
                                                                 galaxy_light_sersic):
        grids.image = np.array([[1.0, 1.0],
                                [1.0, 0.0],
                                [-1.0, 0.0]])

        galaxy_intensities = galaxy_light_sersic.intensity_from_grid(grids.image)

        tracer_intensities = ray_tracing.intensities_via_grid(image_grid=grids.image,
                                                              galaxies=[galaxy_light_sersic, galaxy_light_sersic,
                                                                        galaxy_light_sersic])

        assert (3.0 * galaxy_intensities == tracer_intensities).all()


class TestIntensitiesViaSubGrid:

    def test__no_galaxies__intensities_returned_as_0s(self, grids, galaxy_no_profiles):
        intensities = ray_tracing.intensities_via_sub_grid(sub_grid=grids.sub,
                                                           galaxies=[galaxy_no_profiles])

        assert intensities[0] == 0.0

    def test__galaxy_light_sersic__intensities_returned_as_correct_values(self, grids,
                                                                          galaxy_light_sersic):
        galaxy_intensity = galaxy_light_sersic.intensity_from_grid(grids.sub)

        galaxy_intensity = (galaxy_intensity[0] + galaxy_intensity[1] + galaxy_intensity[2] +
                            galaxy_intensity[3]) / 4.0

        tracer_intensities = ray_tracing.intensities_via_sub_grid(sub_grid=grids.sub,
                                                                  galaxies=[galaxy_light_sersic])

        assert tracer_intensities[0] == galaxy_intensity

    def test__galaxy_light_sersic_x3__deflections_tripled_from_above(self, grids,
                                                                     galaxy_light_sersic):
        galaxy_intensity = galaxy_light_sersic.intensity_from_grid(grids.sub)

        galaxy_intensity = (galaxy_intensity[0] + galaxy_intensity[1] + galaxy_intensity[2] +
                            galaxy_intensity[3]) / 4.0

        tracer_intensities = ray_tracing.intensities_via_sub_grid(sub_grid=grids.sub,
                                                                  galaxies=[galaxy_light_sersic, galaxy_light_sersic,
                                                                            galaxy_light_sersic])

        assert tracer_intensities[0] == pytest.approx(3.0 * galaxy_intensity, 1e-4)


class TestSetupTracedGrid:

    def test__simple_sis_model__deflection_angles(self, grids, galaxy_mass_sis):
        deflections = ray_tracing.deflections_for_grids(grids, [galaxy_mass_sis])

        grid_traced = ray_tracing.traced_collection_for_deflections(grids, deflections)

        assert grid_traced.image[0] == pytest.approx(np.array([1.0 - 0.707, 1.0 - 0.707]), 1e-2)

    def test_three_identical_lenses__deflection_angles_triple(self, grids, galaxy_mass_sis):
        deflections = ray_tracing.deflections_for_grids(grids, [galaxy_mass_sis,
                                                                galaxy_mass_sis,
                                                                galaxy_mass_sis])

        grid_traced = ray_tracing.traced_collection_for_deflections(grids, deflections)

        assert grid_traced.image[0] == pytest.approx(np.array([1.0 - 3.0 * 0.707, 1.0 - 3.0 * 0.707]), 1e-3)

    def test_one_lens_with_three_identical_mass_profiles__deflection_angles_triple(self, grids,
                                                                                   lens_sis_x3):
        deflections = ray_tracing.deflections_for_grids(grids, [lens_sis_x3])

        grid_traced = ray_tracing.traced_collection_for_deflections(grids, deflections)

        assert grid_traced.image[0] == pytest.approx(np.array([1.0 - 3.0 * 0.707, 1.0 - 3.0 * 0.707]), 1e-3)



import os
import pickle

from autolens.imaging import scaled_array
from autolens.imaging import image
from autolens.imaging import mask
from autolens.imaging import masked_image

path =  "{}/".format(os.path.dirname(os.path.realpath(__file__)))

def load_data(name, pixel_scale, psf_shape):
    im = scaled_array.ScaledArray.from_fits(file_path=path + 'data/'+name+'/image', hdu=0, pixel_scale=pixel_scale)
    noise = scaled_array.ScaledArray.from_fits(file_path=path + 'data/'+name+'/noise', hdu=0, pixel_scale=pixel_scale)
    exposure_time = scaled_array.ScaledArray.from_fits(file_path=path + 'data/'+name+'/exposure_time', hdu=0,
                                                       pixel_scale=pixel_scale)
    psf = image.PSF.from_fits(file_path=path + 'data/LSST/psf', hdu=0, pixel_scale=pixel_scale).trim(psf_shape)

    return im, noise, exposure_time, psf

def setup_class(name, pixel_scale, radius_mask=4.0, psf_shape=(21,21), sub_grid_size=4):

    def pickle_path():
        return path + 'data/' + name + '/pickle/r' + str(radius_mask) + '_psf' + str(psf_shape[0]) + '_sub' + \
               str(sub_grid_size)

    if not os.path.isdir(path + 'data/' + name + '/pickle'):
        os.mkdir(path + 'data/' + name + '/pickle')

    if not os.path.isfile(pickle_path()):
        return Data(name, pixel_scale, radius_mask, psf_shape, sub_grid_size)
    elif os.path.isfile(pickle_path()):
        with open(pickle_path(), 'rb') as pickle_file:
             thing=pickle.load(file=pickle_file)
        return thing


class Data(object):

    def __init__(self, name, pixel_scale, radius_mask=4.0, psf_shape=(21, 21), sub_grid_size=4):

        def pickle_path():
            return path + 'data/' + name + '/pickle/r' + str(radius_mask) + '_psf' + str(psf_shape[0]) + '_sub' + \
                   str(sub_grid_size)

        im, noise, exposure_time, psf = load_data(name=name, pixel_scale=pixel_scale, psf_shape=psf_shape)

        im = image.Image(array=im, effective_exposure_time=exposure_time, pixel_scale=pixel_scale, psf=psf,
                         background_noise=noise, poisson_noise=noise)

        ma = mask.Mask.circular(shape_arc_seconds=im.shape_arc_seconds, pixel_scale=im.pixel_scale,
                                radius_mask=radius_mask)

        self.masked_image = masked_image.MaskedImage(image=im, mask=ma)

        self.grids = mask.GridCollection.from_mask_sub_grid_size_and_blurring_shape(mask=ma,
                                                                                   sub_grid_size=sub_grid_size,
                                                                                   blurring_shape=psf.shape)

        self.borders = mask.BorderCollection.from_mask_and_sub_grid_size(mask=ma, sub_grid_size=sub_grid_size)

        with open(pickle_path(), 'wb') as pickle_file:
            pickle.dump(self, file=pickle_file)
